#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2010, 2011, 2012, 2013, 2014, 2015, 2017.

# SMHI,
# Folkborgsvägen 1,
# Norrköping,
# Sweden

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>
#   Adam Dybbroe <adam.dybbroe@smhi.se>

# This file is part of mpop.

# mpop is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# mpop is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# mpop.  If not, see <http://www.gnu.org/licenses/>.

"""This module defines satellite instrument channels as a generic class, to be
inherited when needed.
"""
import copy

import numpy as np
import logging

LOG = logging.getLogger(__name__)

try:
    from pyorbital.astronomy import sun_zenith_angle as sza
except ImportError:
    sza = None

from mpop.tools import viewzen_corr as vz_corr


class GeolocationIncompleteError(Exception):

    """Exception to try catch cases where the original data have not been read or
    expanded properly so that each pixel has a geo-location"""

    pass


class NotLoadedError(Exception):

    """Exception to be raised when attempting to use a non-loaded channel.
    """
    pass


class GenericChannel(object):

    """This is an abstract channel class. It can be a super class for
    calibrated channels data or more elaborate channels such as cloudtype or
    CTTH.
    """

    def __init__(self, name=None):
        object.__init__(self)

        # Channel name
        if name is not None and not isinstance(name, str):
            raise TypeError("Channel name must be a string, or None")
        self.name = name

        # Channel resolution, in meters.
        self.resolution = None

        # ID of the area on which the channel is defined.
        self.area_id = None

        # Area on which the channel is defined.
        self.area_def = None
        self.info = {}

    def __cmp__(self, ch2):
        if(isinstance(ch2, str)):
            return cmp(self.name, ch2)
        elif(ch2.name is not None and
             self.name is not None and
             ch2.name[0] == "_" and
             self.name[0] != "_"):
            return -1
        elif(ch2.name is not None and
             self.name is not None and
             ch2.name[0] != "_" and
             self.name[0] == "_"):
            return 1
        else:
            return cmp(self.name, ch2.name)

    def _get_area(self):
        """Getter for area.
        """
        return self.area_def or self.area_id

    def _set_area(self, area):
        """Setter for area.
        """
        if (area is None):
            self.area_def = None
            self.area_id = None
        elif(isinstance(area, str)):
            self.area_id = area
        else:
            try:
                dummy = area.area_extent
                dummy = area.x_size
                dummy = area.y_size
                dummy = area.proj_id
                dummy = area.proj_dict
                self.area_def = area
            except AttributeError:
                try:
                    dummy = area.lons
                    dummy = area.lats
                    self.area_def = area
                    self.area_id = None
                except AttributeError:
                    raise TypeError("Malformed area argument. "
                                    "Should be a string or an area object.")

    area = property(_get_area, _set_area)


class Channel(GenericChannel):

    """This is the satellite channel class. It defines satellite channels as a
    container for calibrated channel data.

    The *resolution* sets the resolution of the channel, in meters. The
    *wavelength_range* is a triplet, containing the lowest-, center-, and
    highest-wavelength values of the channel. *name* is simply the given name
    of the channel, and *data* is the data it should hold.
    """

    def __init__(self,
                 name=None,
                 resolution=0,
                 wavelength_range=[-np.inf, -np.inf, -np.inf],
                 data=None,
                 calibration_unit=None):

        GenericChannel.__init__(self, name)

        self._data = None
        self.wavelength_range = None

        if(name is None and
           wavelength_range == [-np.inf, -np.inf, -np.inf]):
            raise ValueError("Cannot define a channel with neither name "
                             "nor wavelength range.")

        if not isinstance(resolution, (int, float)):
            raise TypeError("Resolution must be an integer number of meters.")

        self.resolution = resolution

        if(not isinstance(wavelength_range, (tuple, list, set)) or
           len(wavelength_range) != 3 or
           not isinstance(wavelength_range[0], float) or
           not isinstance(wavelength_range[1], float) or
           not isinstance(wavelength_range[2], float)):
            raise TypeError("Wavelength_range should be a triplet of floats.")
        elif(not (wavelength_range[0] <= wavelength_range[1]) or
             not (wavelength_range[1] <= wavelength_range[2])):
            raise ValueError("Wavelength_range should be a sorted triplet.")

        self.wavelength_range = list(wavelength_range)
        self.unit = calibration_unit
        self.data = data

    def get_reflectance(self, tb11, sun_zenith=None, tb13_4=None):
        """Get the reflectance part of an NIR channel"""

        try:
            from pyspectral.near_infrared_reflectance import Calculator
        except ImportError:
            LOG.info("Couldn't load pyspectral")

        # Check the wavelength, and if outside 3-4 microns this functionality
        # doesn't give any meaning and should not be supported
        if (self.wavelength_range[1] < 3.0 or self.wavelength_range[1] > 4.0):
            LOG.warning("Deriving the near infrared reflectance" +
                        " of a band that is outside the 3-4 micron range" +
                        " is not supported!\n\tWill do nothing...")
            return

        # Check if the sun-zenith angle was provided:
        if sun_zenith is None:
            lonlats = self.area.get_lonlats()
            sun_zenith = sza(self.info['time'], lonlats[0], lonlats[1])

        try:
            refl39 = Calculator(self.info['satname'] + self.info['satnumber'],
                                self.info['instrument_name'], self.name)
        except NameError:
            LOG.warning("pyspectral missing!")
            return

        return refl39.reflectance_from_tbs(sun_zenith, self.data,
                                           tb11, tb13_4)

    def __cmp__(self, ch2, key=0):
        if(isinstance(ch2, str)):
            return cmp(self.name, ch2)
        elif(ch2.name is not None and
             self.name is not None and
             ch2.name[0] == "_" and
             self.name[0] != "_"):
            return -1
        elif(ch2.name is not None and
             self.name is not None and
             ch2.name[0] != "_" and
             self.name[0] == "_"):
            return 1
        else:
            res = cmp(abs(self.wavelength_range[1] - key),
                      abs(ch2.wavelength_range[1] - key))
            if res == 0:
                return cmp(self.name, ch2.name)
            else:
                return res

    def __str__(self):
        if self.shape is not None:
            return ("'%s: (%.3f,%.3f,%.3f)μm, shape %s, resolution %sm'" %
                    (self.name,
                     self.wavelength_range[0],
                     self.wavelength_range[1],
                     self.wavelength_range[2],
                     self.shape,
                     self.resolution))
        else:
            return ("'%s: (%.3f,%.3f,%.3f)μm, resolution %sm, not loaded'" %
                    (self.name,
                     self.wavelength_range[0],
                     self.wavelength_range[1],
                     self.wavelength_range[2],
                     self.resolution))

    def is_loaded(self):
        """Tells if the channel contains loaded data.
        """
        return self._data is not None

    def check_range(self, min_range=1.0):
        """Check that the data of the channels has a definition domain broader
        than *min_range* and return the data, otherwise return zeros.
        """
        if not self.is_loaded():
            raise ValueError("Cannot check range of an non-loaded channel")

        if not isinstance(min_range, (float, int)):
            raise TypeError("Min_range must be a single number.")

        if isinstance(self._data, np.ma.core.MaskedArray):
            if self._data.mask.all():
                return self._data

        if((self._data.max() - self._data.min()) < min_range):
            return np.ma.zeros(self.shape)
        else:
            return self._data

    def show(self):
        """Display the channel as an image.
        """
        if not self.is_loaded():
            raise ValueError("Channel not loaded, cannot display.")

        from PIL import Image as pil

        data = ((self._data - self._data.min()) * 255.0 /
                (self._data.max() - self._data.min()))
        if isinstance(data, np.ma.core.MaskedArray):
            img = pil.fromarray(np.array(data.filled(0), np.uint8))
        else:
            img = pil.fromarray(np.array(data, np.uint8))
        img.show()

    def as_image(self, stretched=True):
        """Return the channel as a :class:`mpop.imageo.geo_image.GeoImage`
        object. The *stretched* argument set to False allows the data to remain
        untouched (as opposed to crude stretched by default to obtain the same
        output as :meth:`show`).
        """
        from mpop.imageo.geo_image import GeoImage

        img = GeoImage(self._data, self.area, None)
        if stretched:
            img.stretch("crude")
        return img

    def project(self, coverage_instance):
        """Make a projected copy of the current channel using the given
        *coverage_instance*.

        See also the :mod:`mpop.projector` module.
        """
        res = Channel(name=self.name,
                      resolution=self.resolution,
                      wavelength_range=self.wavelength_range,
                      data=None,
                      calibration_unit=self.unit)
        res.area = coverage_instance.out_area
        res.info = self.info
        if hasattr(self, 'palette'):      # UH, new
            res.palette = self.palette    # UH, new
        if self.is_loaded():
            LOG.info("Projecting channel %s (%fμm)..."
                     % (self.name, self.wavelength_range[1]))
            import pyresample
            if (hasattr(coverage_instance, 'in_area') and
                isinstance(coverage_instance.in_area, pyresample.geometry.SwathDefinition) and
                    hasattr(coverage_instance.in_area.lats, 'shape') and
                    coverage_instance.in_area.lats.shape != self._data.shape):
                raise GeolocationIncompleteError("Lons and lats doesn't match data! " +
                                                 "Data can't be re-projected unless " +
                                                 "each pixel of the swath has a " +
                                                 "geo-location atached to it.")
            data = coverage_instance.project_array(self._data)
            res.data = data
            return res
        else:
            raise NotLoadedError("Can't project, channel %s (%fμm) not loaded."
                                 % (self.name, self.wavelength_range[1]))

    def get_data(self):
        """Getter for channel data.
        """
        return self._data

    def set_data(self, data):
        """Setter for channel data.
        """
        if data is None:
            del self._data
            self._data = None
        elif isinstance(data, (np.ndarray, np.ma.core.MaskedArray)):
            self._data = data
        else:
            raise TypeError("Data must be a numpy (masked) array.")

    data = property(get_data, set_data)

    @property
    def shape(self):
        """Shape of the channel.
        """
        if self.data is None:
            return None
        else:
            return self.data.shape

    def sunzen_corr(self, time_slot, lonlats=None, limit=80., mode='cos',
                    sunmask=False):
        '''Perform Sun zenith angle correction for the channel at
        *time_slot* (datetime.datetime() object) and return the
        corrected channel.  The parameter *limit* can be used to set
        the maximum zenith angle for which the correction is
        calculated.  For larger angles, the correction is the same as
        at the *limit* (default: 80.0 degrees).  Coordinate values can
        be given as a 2-tuple or a two-element list *lonlats* of numpy
        arrays; if None, the coordinates will be read from the channel
        data.  Parameter *mode* is a placeholder for other possible
        illumination corrections. The name of the new channel will be
        *original_chan.name+'_SZC'*, eg. "VIS006_SZC".  This name is
        also stored to the info dictionary of the originating channel.
        '''

        if self.info.get('sun_zen_correction_applied'):
            LOG.debug("Sun zenith correction already applied, skipping")
            return self

        import mpop.tools

        try:
            from pyorbital import astronomy
        except ImportError:
            LOG.warning("Could not load pyorbital.astronomy")
            return None

        if lonlats is None or len(lonlats) != 2:
            # Read coordinates
            LOG.debug("No valid coordinates given, reading from the "
                      "channel data")
            lons, lats = self.area.get_lonlats()
        else:
            lons, lats = lonlats

        # Calculate Sun zenith angles and the cosine
        cos_zen = astronomy.cos_zen(time_slot, lons, lats)

        # Copy the channel
        new_ch = copy.deepcopy(self)

        # Set the name
        new_ch.name += '_SZC'

        if mode == 'cos':
            new_ch.data = mpop.tools.sunzen_corr_cos(new_ch.data,
                                                     cos_zen, limit=limit)
        else:
            # Placeholder for other correction methods
            pass

        # Add information about the corrected version to original
        # channel
        self.info["sun_zen_corrected"] = self.name + '_SZC'

        if sunmask:
            if isinstance(sunmask, (float, int)):
                sunmask = sunmask
            else:
                sunmask = 90.
            cos_limit = np.cos(np.radians(sunmask))
            LOG.debug("Masking out data where sun-zenith " +
                      "is greater than %f deg", sunmask)
            LOG.debug("cos_limit = %f", cos_limit)
            # Mask out data where the sun elevation is below a threshold:
            new_ch.data = np.ma.masked_where(
                cos_zen < cos_limit, new_ch.data, copy=False)

        new_ch.info["sun_zen_correction_applied"] = True

        return new_ch

    def get_viewing_geometry(self, orbital, time_slot, altitude=None):
        '''Calculates the azimuth and elevation angle as seen by the observer 
           at the position of the current area pixel. 
           inputs:
             orbital   an orbital object define by the tle file 
                       (see pyorbital.orbital import Orbital or mpop/scene.py get_oribtal)
             time_slot time object specifying the observation time
             altitude  optinal: altitude of the observer above the earth ellipsoid
           outputs:
             azi       azimuth viewing angle in degree (south is 0, counting clockwise)
             ele       elevation viewing angle in degree (zenith is 90, horizon is 0)
        '''

        try:
            from pyorbital.orbital import Orbital
        except ImportError:
            LOG.warning("Could not load pyorbital.orbial.Orbital")
            return None

        try:
            from pyorbital import tlefile
        except ImportError:
            LOG.warning("Could not load pyorbital.tlefile")
            return None

        (lons, lats) = self.area.get_lonlats()
        # Calculate observer azimuth and elevation
        if altitude == None:
            altitude = np.zeros(lons.shape)
        azi, ele = orbital.get_observer_look(time_slot, lons, lats, altitude)

        return (azi, ele)

    def vinc_vect(phi, lembda, alpha, s, f=None, a=None, degree=True):
        """ Vincenty's Direct formular

        Returns the lat and long of projected point and reverse azimuth
        given a reference point and a distance and azimuth to project.
        lats, longs and azimuths are passed in radians.

        Keyword arguments:
            phi    Latitude in degree/radians
            lembda Longitude in degree/radians
            alpha    Geodetic azimuth in degree/radians
            s    Ellipsoidal distance in meters
            f    WGS84 parameter
            a    WGS84 parameter
            degree Boolean if in/out values are in degree or radians.
                   Default is in degree

        Returns:
            (phiout,  lembdaout,  alphaout ) as a tuple

        """
        if degree:
            phi = np.deg2rad(phi)
            lembda = np.deg2rad(lembda)
            alpha = np.deg2rad(alpha)

        if f is None:
            f = 1 / 298.257223563
        if a is None:
            a = 6378137

        two_pi = 2.0 * np.pi

        if isinstance(alpha, np.ndarray):
            alpha[alpha < 0.0] += two_pi
            alpha[alpha > two_pi] -= two_pi

        else:
            if alpha < 0.0:
                alpha = alpha + two_pi
            if (alpha > two_pi):
                alpha = alpha - two_pi
        """
        alphama = np.ma.masked_less_equal(alphama, two_pi)
        alpha = alphama - two_pi
        alpha.mask = np.ma.nomask
        logger.debug(alpha)
        """
        b = a * (1.0 - f)

        tan_u1 = (1 - f) * np.tan(phi)
        u_1 = np.arctan(tan_u1)
        sigma1 = np.arctan2(tan_u1, np.cos(alpha))

        sinalpha = np.cos(u_1) * np.sin(alpha)
        cosalpha_sq = 1.0 - sinalpha * sinalpha

        u_2 = cosalpha_sq * (a * a - b * b) / (b * b)
        aa_ = 1.0 + (u_2 / 16384) * (4096 + u_2 * (-768 + u_2 *
                                                   (320 - 175 * u_2)))
        bb_ = (u_2 / 1024) * (256 + u_2 * (-128 + u_2 * (74 - 47 * u_2)))

        # Starting with the approximation
        sigma = (s / (b * aa_))
        last_sigma = 2.0 * sigma + 2.0  # something impossible

        # Iterate the following three equations
        # until there is no significant change in sigma

        # two_sigma_m , delta_sigma

        def iter_sigma(sigma, last_sigma, sigma1, s, b, aa_, bb_):
            while (abs((last_sigma - sigma) / sigma) > 1.0e-9):
                two_sigma_m = 2 * sigma1 + sigma

                delta_sigma = (bb_ * np.sin(sigma) *
                               (np.cos(two_sigma_m) + (bb_ / 4) *
                                (np.cos(sigma) *
                                 (-1 + 2 * np.power(np.cos(two_sigma_m), 2) -
                                  (bb_ / 6) * np.cos(two_sigma_m) *
                                  (-3 + 4 * np.power(np.sin(sigma), 2)) *
                                  (-3 + 4 * np.power(np.cos(two_sigma_m), 2))))))
                last_sigma = sigma
                sigma = (s / (b * aa_)) + delta_sigma

            return(sigma, two_sigma_m)

        # Check for array inputs
        arraybool = [isinstance(ele, np.ndarray) for ele in (sigma, last_sigma,
                                                             sigma1)]
        logger.debug("Sigma Arrays?: " + str(arraybool))
        if all(arraybool):
            viter_sigma = np.vectorize(iter_sigma)
            sigma, two_sigma_m = viter_sigma(sigma, last_sigma, sigma1, s, b, aa_,
                                             bb_)

        else:
            sigma, two_sigma_m = iter_sigma(sigma, last_sigma, sigma1, s, b, aa_,
                                            bb_)

        phiout = np.arctan2((np.sin(u_1) * np.cos(sigma) +
                             np.cos(u_1) * np.sin(sigma) * np.cos(alpha)),
                            ((1 - f) * np.sqrt(np.power(sinalpha, 2) +
                                               pow(np.sin(u_1) *
                                                   np.sin(sigma) -
                                                   np.cos(u_1) *
                                                   np.cos(sigma) *
                                                   np.cos(alpha), 2))))

        deltalembda = np.arctan2((np.sin(sigma) * np.sin(alpha)),
                                 (np.cos(u_1) * np.cos(sigma) -
                                  np.sin(u_1) * np.sin(sigma) * np.cos(alpha)))

        cc_ = (f / 16) * cosalpha_sq * (4 + f * (4 - 3 * cosalpha_sq))

        omega = (deltalembda - (1 - cc_) * f * sinalpha *
                 (sigma + cc_ * np.sin(sigma) * (np.cos(two_sigma_m) + cc_ *
                                                 np.cos(sigma) *
                                                 (-1 + 2 *
                                                  np.power(np.cos(two_sigma_m),
                                                           2)))))

        lembdaout = lembda + omega

        alphaout = np.arctan2(sinalpha, (-np.sin(u_1) * np.sin(sigma) +
                                         np.cos(u_1) * np.cos(sigma) *
                                         np.cos(alpha)))

        alphaout = alphaout + two_pi / 2.0

        if isinstance(alphaout, np.ndarray):
            alphaout[alphaout < 0.0] += two_pi
            alphaout[alphaout > two_pi] -= two_pi

        else:
            if alphaout < 0.0:
                alphaout = alphaout + two_pi
            if (alphaout > two_pi):
                alphaout = alphaout - two_pi

        if degree:
            phiout = np.rad2deg(phiout)
            lembdaout = np.rad2deg(lembdaout)
            alphaout = np.rad2deg(alphaout)

        return(phiout, lembdaout, alphaout)

    def parallax_corr(self, cth=None, time_slot=None, orbital=None, azi=None, ele=None, fill="False"):
        '''Perform the parallax correction for channel at
        *time_slot* (datetime.datetime() object), assuming the cloud top height cth
        and the viewing geometry given by the satellite orbital "orbital" and return the
        corrected channel. 
        Authors: Ulrich Hamann (MeteoSwiss), Thomas Leppelt (DWD)
        Example calls:
            * calling this function (using orbital and time_slot)
                 orbital = data.get_oribtal()
                 data["VIS006"].parallax_corr(cth=data["CTTH"].height, time_slot=data.time_slot, orbital=orbital)
            * calling this function (using viewing geometry)
                 orbital = data.get_oribtal()
                 (azi, ele) = get_viewing_geometry(self, orbital, time_slot)
                 data["VIS006"].parallax_corr(cth=data["CTTH"].height, azi=azi, ele=ele)
        Optional input:
          cth        The parameter cth is the cloud top height 
                     (or  the altitude of the object that should be shifted).
                     cth must have the same size and projection as the channel

          orbital    an orbital object define by the tle file 
                     (see pyorbital.orbital import Orbital or mpop/scene.py get_oribtal)
          azi        azimuth viewing angle in degree (south is 0, counting clockwise)
                     e.g. as given by self.get_viewing_geometry
          ele        elevation viewing angle in degree (zenith is 90, horizon is 0)
                     e.g. as given by self.get_viewing_geometry
          fill       specifies the interpolation method to fill the gaps
                     (basically areas behind the cloud that can't be observed by the satellite instrument)
                     "False" (default): no interpolation, gaps are np.nan values and mask is set accordingly
                     "nearest": fill gaps with nearest neighbour
                     "bilinear": use scipy.interpolate.griddata with linear interpolation 
                                 to fill the gaps

        output: 
          parallax corrected channel
                     the content of the channel will be parallax corrected.
                     The name of the new channel will be
                     *original_chan.name+'_PC'*, eg. "IR_108_PC". This name is
                     also stored to the info dictionary of the originating channel.
        '''

        # get time_slot from info, if present
        if time_slot == None:
            if "time" in self.info.keys():
                time_slot = self.info["time"]

        if azi == None or ele == None:
            if time_slot == None or orbital == None:
                print "*** Error in parallax_corr (mpop/channel.py)"
                print "    parallax_corr needs either time_slot and orbital"
                print "    data[\"IR_108\"].parallax_corr(data[\"CTTH\"].height, time_slot=data.time_slot, orbital=orbital)"
                print "    or the azimuth and elevation angle"
                print "    data[\"IR_108\"].parallax_corr(data[\"CTTH\"].height, azi=azi, ele=ele)"
                quit()
            else:
                print (
                    "... calculate viewing geometry (orbit and time are given)")
                (azi, ele) = self.get_viewing_geometry(orbital, time_slot)
        else:
            print ("... azimuth and elevation angle given")

        # mask the cloud top height
        cth_ = np.ma.masked_where(cth < 0, cth, copy=False)

        # Elevation displacement
        dz = cth_ / np.tan(np.deg2rad(ele))

        # Create the new channel (by copying) and initialize the data with None
        # values
        new_ch = copy.deepcopy(self)
        new_ch.data[:, :] = np.nan

        # Set the name
        new_ch.name += '_PC'

        # Add information about the corrected version to original channel
        self.info["parallax_corrected"] = self.name + '_PC'

        # get projection coordinates in meter
        (proj_x, proj_y) = self.area.get_proj_coords()

        print "... calculate parallax shift"
        # shifting pixels according to parallax corretion
        # shift West-East   in m  # ??? sign correct ???
        proj_x_pc = proj_x - np.sin(np.deg2rad(azi)) * dz
        # shift North-South in m
        proj_y_pc = proj_y + np.cos(np.deg2rad(azi)) * dz

        # get indices for the pixels for the original position
        (y, x) = self.area.get_xy_from_proj_coords(proj_x, proj_y)
        # comment: might be done more efficient with meshgrid
        # >>> x = np.arange(-5.01, 5.01, 0.25)
        # >>> y = np.arange(-5.01, 5.01, 0.25)
        # >>> xx, yy = np.meshgrid(x, y)
        # get indices for the pixels at the parallax corrected position
        (y_pc, x_pc) = self.area.get_xy_from_proj_coords(proj_x_pc, proj_y_pc)

        # copy cloud free satellite pixels (surface observations)
        ind = np.where(cth_.mask == True)
        new_ch.data[x[ind], y[ind]] = self.data[x[ind], y[ind]]

        print "... copy data to parallax corrected position"
        # copy cloudy pixel with new position modified with parallax shift
        ind = np.where(x_pc.mask == False)
        new_ch.data[x_pc[ind], y_pc[ind]] = self.data[x[ind], y[ind]]

        # Mask out data gaps (areas behind the clouds)
        new_ch.data = np.ma.masked_where(
            np.isnan(new_ch.data), new_ch.data, copy=False)

        if fill.lower() == "false":
            return new_ch
        elif fill == "nearest":
            print "*** fill missing values with nearest neighbour"
            from scipy.ndimage import distance_transform_edt
            invalid = np.isnan(new_ch.data)
            ind = distance_transform_edt(
                invalid, return_distances=False, return_indices=True)
            new_ch.data = new_ch.data[tuple(ind)]
        elif fill == "bilinear":
            # this function does not interpolate at the outer boundaries
            from scipy.interpolate import griddata
            ind = np.where(new_ch.data.mask == False)
            points = np.transpose(np.append([y[ind]], [x[ind]], axis=0))
            values = new_ch.data[ind]
            new_ch.data = griddata(points, values, (y, x), method='linear')

            # fill the remaining pixels with nearest neighbour
            from scipy.ndimage import distance_transform_edt
            invalid = np.isnan(new_ch.data)
            ind = distance_transform_edt(
                invalid, return_distances=False, return_indices=True)
            new_ch.data = new_ch.data[tuple(ind)]
        else:
            print "*** Error in parallax_corr (channel.py)"
            print "    unknown gap fill method ", fill
            quit()

        return new_ch

    def viewzen_corr(self, view_zen_angle_data):
        """Apply atmospheric correction on a copy of this channel data
        using the given satellite zenith angle data of the same shape.
        Returns a new channel containing the corrected data.
        The name of the new channel will be *original_chan.name+'_VZC'*,
        eg. "IR108_VZC".  This name is also stored to the info dictionary of
        the originating channel.
        """

        # copy channel data which will be corrected in place
        chn_data = self.data.copy()
        CHUNK_SZ = 500
        for start in xrange(0, chn_data.shape[1], CHUNK_SZ):
            # apply correction on channel data
            vz_corr(chn_data[:, start:start + CHUNK_SZ],
                    view_zen_angle_data[:, start:start + CHUNK_SZ])

        new_ch = Channel(name=self.name + "_VZC",
                         resolution=self.resolution,
                         wavelength_range=self.wavelength_range,
                         data=chn_data,
                         calibration_unit=self.unit)

        # Add information about the corrected version to original channel
        self.info["view_zen_corrected"] = self.name + '_VZC'

        return new_ch

    # Arithmetic operations on channels.

    def __pow__(self, other):
        return Channel(name="new", data=self.data ** other)

    def __rpow__(self, other):
        return Channel(name="new", data=self.data ** other)

    def __mul__(self, other):
        return Channel(name="new", data=self.data * other)

    def __rmul__(self, other):
        return Channel(name="new", data=self.data * other)

    def __add__(self, other):
        return Channel(name="new", data=self.data + other)

    def __radd__(self, other):
        return Channel(name="new", data=self.data + other)

    def __sub__(self, other):
        return Channel(name="new", data=self.data - other)

    def __rsub__(self, other):
        return Channel(name="new", data=self.data - other)

    def __div__(self, other):
        return Channel(name="new", data=self.data / other)

    def __rdiv__(self, other):
        return Channel(name="new", data=self.data / other)

    def __neg__(self):
        return Channel(name="new", data=-self.data)

    def __abs__(self):
        return Channel(name="new", data=abs(self.data))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2010, 2012.

# SMHI,
# Folkborgsvägen 1,
# Norrköping, 
# Sweden

# Author(s):
 
#   Martin Raspaud <martin.raspaud@smhi.se>
#   Esben S. Nielsen <esn@dmi.dk>

# This file is part of mpop.

# mpop is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# mpop is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# mpop.  If not, see <http://www.gnu.org/licenses/>.

"""Plugin for reading PPS's cloud products hdf files.
"""
import ConfigParser
import datetime
import os.path
from glob import glob

import mpop.channel
from mpop import CONFIG_PATH
from mpop.utils import get_logger


LOG = get_logger('satin/pps_hdf')

class PpsCloudType(mpop.channel.GenericChannel):
    def __init__(self):
        mpop.channel.GenericChannel.__init__(self, "CloudType")
        self.region = None
        self.des = ""
        self.cloudtype_des = ""
        self.qualityflag_des = ""
        self.phaseflag_des = ""
        self.sec_1970 = 0
        self.satellite_id = ""
        self.cloudtype_lut = []
        self.qualityflag_lut = []
        self.phaseflag_lut = []
        self.cloudtype = None
        self.qualityflag = None
        self.phaseflag = None

    def copy(self, other):
        self.region = other.region
        self.des = other.des
        self.cloudtype_des = other.cloudtype_des
        self.qualityflag_des = other.qualityflag_des
        self.phaseflag_des = other.phaseflag_des
        self.sec_1970 = other.sec_1970
        self.satellite_id = other.satellite_id
        self.cloudtype_lut = other.cloudtype_lut
        self.qualityflag_lut = other.qualityflag_lut
        self.phaseflag_lut = other.phaseflag_lut
        self.cloudtype = other.cloudtype
        self.qualityflag = other.qualityflag
        self.phaseflag = other.phaseflag
        
    def read(self, filename):
        import epshdf
        self.copy(epshdf.read_cloudtype(filename))
        
    def is_loaded(self):
        return self.cloudtype is not None


class PpsCTTH(mpop.channel.GenericChannel):
    def __init__(self):
        mpop.channel.GenericChannel.__init__(self, "CTTH")
        self.region = None
        self.des = ""
        self.ctt_des = ""
        self.cth_des = ""
        self.ctp_des = ""
        self.cloudiness_des = ""
        self.processingflag_des = ""
        self.sec_1970 = 0
        self.satellite_id = ""
        self.processingflag_lut = []

        self.temperature = None
        self.t_gain = 1.0
        self.t_intercept = 0.0
        self.t_nodata = 255

        self.pressure = None
        self.p_gain = 1.0
        self.p_intercept = 0.0
        self.p_nodata = 255

        self.height = None
        self.h_gain = 1.0
        self.h_intercept = 0.0
        self.h_nodata = 255

        self.cloudiness = None
        self.c_nodata = 255
        self.processingflag = None

    def copy(self, other):
        self.region = other.region
        self.des = other.des
        self.ctt_des = other.ctt_des
        self.cth_des = other.cth_des
        self.ctp_des = other.ctp_des
        self.cloudiness_des = other.cloudiness_des
        self.processingflag_des = other.processingflag_des
        self.sec_1970 = other.sec_1970
        self.satellite_id = other.satellite_id
        self.processingflag_lut = other.processingflag_lut
        
        self.temperature = other.temperature
        self.t_gain = other.t_gain
        self.t_intercept = other.t_intercept
        self.t_nodata = other.t_nodata
        
        self.pressure = other.pressure
        self.p_gain = other.p_gain
        self.p_intercept = other.p_intercept
        self.p_nodata = other.p_nodata
        
        self.height = other.height
        self.h_gain = other.h_gain
        self.h_intercept = other.h_intercept
        self.h_nodata = other.h_nodata
        
        self.cloudiness = other.cloudiness
        self.c_nodata = other.c_nodata
        self.processingflag = other.processingflag
        
    def read(self, filename):
        import epshdf
        self.copy(epshdf.read_cloudtop(filename))

def load(scene, **kwargs):
    """Load data into the *channels*. *Channels* is a list or a tuple
    containing channels we will load data into. If None, all channels are
    loaded.
    """

    del kwargs

    if("CTTH" not in scene.channels_to_load and
       "CloudType" not in scene.channels_to_load):
        return
    
    conf = ConfigParser.ConfigParser()
    conf.read(os.path.join(CONFIG_PATH, scene.fullname+".cfg"))
    directory = conf.get(scene.instrument_name+"-level3", "dir")
    filename = conf.get(scene.instrument_name+"-level3", "filename",
                        raw=True)
    pathname = os.path.join(directory, filename)

    area_name = scene.area_id or scene.area.area_id

    times = [scene.time_slot,
             scene.time_slot + datetime.timedelta(minutes=1),
             scene.time_slot - datetime.timedelta(minutes=1)]
    
    if "CTTH" in scene.channels_to_load:
        for time_slot in times:
            filename = (time_slot.strftime(pathname)
                        %{"orbit": scene.orbit,
                          "number": scene.number,
                          "area": area_name,
                          "satellite": scene.fullname,
                          "product": "ctth"})
            flist = glob(filename)
            if len(flist) == 0:
                LOG.info("Can't find " + filename)
            elif len(flist) > 1:
                LOG.info("Too many files matching! " + str(flist))
                break
            else:
                filename = flist[0]
                break
        if not os.path.exists(filename):
            LOG.info("Can't find any CTTH file, skipping")
        else:
            ct_chan = PpsCTTH()
            ct_chan.read(filename)
            ct_chan.area = scene.area
            scene.channels.append(ct_chan)

    if "CloudType" in scene.channels_to_load:
        for time_slot in times:
            filename = (time_slot.strftime(pathname)
                        %{"orbit": scene.orbit,
                          "number": scene.number,
                          "area": area_name,
                          "satellite": scene.fullname,
                          "product": "cloudtype"})
            flist = glob(filename)
            if len(flist) == 0:
                LOG.info("Can't find " + filename)
            elif len(flist) > 1:
                LOG.info("Too many files matching! " + str(flist))
                break
            else:
                filename = flist[0]
                break
        if not os.path.exists(filename):
            LOG.info("Can't find any Cloudtype file, skipping")
        else:
            ct_chan = PpsCloudType()
            ct_chan.read(filename)
            ct_chan.area = scene.area
            scene.channels.append(ct_chan)

    LOG.info("Loading channels done.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2010, 2011, 2012, 2014, 2016.

# Author(s):

#   Kristian Rune Larssen <krl@dmi.dk>
#   Adam Dybbroe <adam.dybbroe@smhi.se>
#   Martin Raspaud <martin.raspaud@smhi.se>
#   Esben S. Nielsen <esn@dmi.dk>

# This file is part of mpop.

# mpop is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# mpop is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# mpop.  If not, see <http://www.gnu.org/licenses/>.


"""mpop netcdf4 writer interface.
"""

__revision__ = 0.1

import numpy as np
import logging

logger = logging.getLogger(__name__)


def save(scene, filename, compression=True, dtype=np.int16, band_axis=2,
         area_aggregation=True, time_dimension=False):
    """Saves the scene as a NetCDF4 file, with CF conventions.

    *band_axis* gives which axis to use for the band dimension. For
     example, use band_axis=0 to get dimensions like (band, y, x).

    *area_aggregation* determines if bands on the same area should be gathered
     together or not. Default is True, meaning aggregation. If
     *area_aggregation* is False, the band_axis gets obsolete. Area aggregation
     is currently not possible when using *time_dimension*.

    *time_dimension* is a boolean and if True a time axis (dimension=1) is
    added in front, like (time, y, x), and the band_axis is omitted. Thus each
    data/band go in a separate dataset.

    """
    from mpop.satout.cfscene import CFScene

    scene.add_to_history("Saved as netcdf4/cf by pytroll/mpop.")
    return netcdf_cf_writer(filename,
                            CFScene(scene, dtype, band_axis, area_aggregation,
                                    time_dimension),
                            compression=compression)


class WriterDimensionError(Exception):

    """ Basic writer exception """
    pass


def attribute_dispenser(info):
    """ returns valid attribute key value pairs
    (for cosmetic reasons, sorted is better than random)"""
    for k, v in sorted(info.iteritems()):
        if k.startswith('var_'):
            continue
        yield (k, v)


def variable_dispenser(root_object, object_list):
    """ Assembles a list of meta info objects """
    # Handle members with info objects
    for v in dir(root_object):
        obj = getattr(root_object, v)
        if callable(obj):
            continue

        # Try to find members with 'info' attribute defined
        # if in list members search through list to find
        # elements with the 'info' attribute defined
        try:
            # test for info attribute on list elements
            for under_obj in obj:
                try:
                    under_obj.info
                    variable_dispenser(under_obj, object_list)
                except AttributeError:
                    pass
        except TypeError:
            try:
                # test for info attribute scalar members
                obj.info
                variable_dispenser(obj, object_list)

            except AttributeError:
                pass

    # Handle local info objects
    try:
        # handle output of member variables without info attribute
        if 'var_children' in root_object.info:
            object_list.extend(root_object.info['var_children'])

        # handle object with info attribute
        object_list.append(root_object.info)
    except AttributeError:
        pass


def find_tag(info_list, tag):
    """ 
        Iterates through info objects to find specific tag. 
        Returns list of matching values.
    """
    tag_data = []
    for info in info_list:
        try:
            tag_data.append(info[tag])
        except KeyError:
            pass
    return tag_data


def find_FillValue_tags(info_list):
    """ 
        Iterates through info objects to find _FillValue tags for var_names
    """
    fill_value_dict = {}
    for info in info_list:
        try:
            fill_value_dict[info['var_name']] = info['_FillValue']
        except KeyError:
            pass
            try:
                fill_value_dict[info['var_name']] = None
            except KeyError:
                pass
    return fill_value_dict


def find_info(info_list, tag):
    """ 
        Iterates through info objects to find specific tag.
        Return list of matching info objects.
    """
    tag_info_objects = []
    for info in info_list:
        if tag in info:
            tag_info_objects.append(info)
    return tag_info_objects


def dtype(element):
    """
        Return the dtype of an array or the type of the element.
    """

    if hasattr(element, "dtype"):
        return element.dtype
    else:
        return type(element)


def shape(element):
    """
        Return the shape of an array or empty tuple if not an array.
    """

    if hasattr(element, "shape"):
        return element.shape
    else:
        return ()


def netcdf_cf_writer(filename, root_object, compression=True):
    """ Write data to file to netcdf file. """
    from netCDF4 import Dataset

    rootgrp = Dataset(filename, 'w')
    try:
        info_list = []
        variable_dispenser(root_object, info_list)

        # find available dimensions
        dim_names = find_tag(info_list, 'var_dim_names')

        # go through all cases of 'var_callback' and create objects which are
        # linked to by the 'var_data' keyword. This ensures that data are only
        # read in when needed.

        cb_infos = find_info(info_list, 'var_callback')

        for info in cb_infos:
            # execute the callback functors
            info['var_data'] = info['var_callback']()

        var_data = find_tag(info_list, 'var_data')

        # create dimensions in NetCDF file, dimension lengths are based on
        # array sizes
        used_dim_names = {}
        for names, values in zip(dim_names, [shape(v) for v in var_data]):

            # case of a scalar
            if len(names) == 0:
                continue
            for dim_name, dim_size in zip(names, values):

                # ensure unique dimension names
                if dim_name in used_dim_names:
                    if dim_size != used_dim_names[dim_name]:
                        raise WriterDimensionError("Dimension name "
                                                   + dim_name +
                                                   " already in use")
                    else:
                        continue

                rootgrp.createDimension(dim_name, dim_size)
                used_dim_names[dim_name] = dim_size

        # create variables

        var_names = find_tag(info_list, 'var_name')

        nc_vars = []

        fill_value_dict = find_FillValue_tags(info_list)
        for name, vtype, dim_name in zip(var_names,
                                         [dtype(vt) for vt in var_data],
                                         dim_names):

            # in the case of arrays containing strings:
            if str(vtype) == "object":
                vtype = str

            nc_vars.append(rootgrp.createVariable(
                name, vtype, dim_name,
                zlib=compression,
                fill_value=fill_value_dict[name]))

        # insert attributes, search through info objects and create global
        # attributes and attributes for each variable.
        for info in info_list:
            if 'var_name' in info:
                # handle variable attributes
                nc_var = rootgrp.variables[info['var_name']]
                nc_var.set_auto_maskandscale(False)
                for j, k in attribute_dispenser(info):
                    if j not in ["_FillValue"]:
                        setattr(nc_var, j, k)
            else:
                # handle global attributes
                for j, k in attribute_dispenser(info):
                    try:
                        setattr(rootgrp, j, k)
                    except TypeError as err:
                        logger.warning("Not saving %s with value %s because %s",
                                       str(j), str(k), str(err))

        # insert data

        for name, vname, vdata in zip(var_names, nc_vars, var_data):
            vname[:] = vdata
    finally:
        rootgrp.close()


if __name__ == '__main__':
    from mpop.satellites.meteosat09 import Meteosat09SeviriScene
    import datetime

    TIME = datetime.datetime(2009, 10, 8, 14, 30)
    GLOB = Meteosat09SeviriScene(area_id="EuropeCanary", time_slot=TIME)
    GLOB.load([0.6, 10.8])

    save(GLOB, 'tester.nc')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright (c) 2015, 2016 Adam.Dybbroe

# Author(s):

#   Adam.Dybbroe <a000680@c20671.ad.smhi.se>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""A reader for the OSISAF SST netCDF format
"""

import os.path
from ConfigParser import ConfigParser

import glob

import mpop.channel
from mpop import CONFIG_PATH
from mpop.plugin_base import Reader

from netCDF4 import Dataset
import numpy as np
import time
from datetime import datetime

import logging
LOG = logging.getLogger(__name__)


class InfoObject(object):

    """Simple data and metadata/header container.
    """

    def __init__(self):
        self.info = {}
        self.data = None


class OsisafNarSstProduct(mpop.channel.GenericChannel):

    def __init__(self, filename=None):
        mpop.channel.GenericChannel.__init__(self)
        self.name = "SST"
        self.mda = {}
        #self.header = {}
        self._projectables = []
        self._keys = []
        self._refs = {}

        self.time = None
        self.dtime = None
        self.sec_1981 = None
        self.sec_1970_1981 = time.mktime(
            (1981, 1, 1, 0, 0, 0, 0, 0, 0)) - time.timezone

        self.file = None
        self.sst = None
        self.l2pf = None  # L2P flags
        self.dt = None  # DT_analysis (K): (SST Deviation from previous day)
        self.stdv = None  # Standard deviation
        self.bias = None  # Bias
        self.lon = None
        self.lat = None

        self.shape = None
        if filename:
            self.read(filename)

    def read(self, filename, load_lonlat=True):
        """Read the OSISAF SST netCDF formatet data (from Ifremer)"""
        LOG.debug("OSISAF netCDF file format...")

        self.file = Dataset(filename, 'r')

        self.fillheader()

        # SST (K):
        sst_data = self.file.variables['sea_surface_temperature']

        sstdata = sst_data[0]
        self.sst = InfoObject()
        # For some strange reason the array seem to start in the lower left!?
        self.sst.data = sstdata[::-1]
        self.sst.info = self.get_data_header(self.sst.info, sst_data)
        self._projectables.append('sst')

        # dtime:
        dtime = self.file.variables['sst_dtime']
        dtime_data = dtime[0] * dtime.scale_factor + dtime.add_offset
        dtime_obj = InfoObject()
        dtime_obj.data = dtime_data[::-1]
        dtime_obj.info = self.get_data_header(dtime_obj.info, dtime)
        self.sec_1981 = dtime_obj.data + self.file.variables['time'][0]
        self.dtime = dtime_obj
        self._projectables.append('dtime')

        # DT_analysis (K): (SST Deviation from previous day)
        dta = self.file.variables['dt_analysis']
        gain = 0.1
        nodata = 255
        offset = -12.7
        data = dta[0] * dta.scale_factor + dta.add_offset
        valid_min = dta.valid_min
        valid_max = dta.valid_max

        dt_data = np.where(np.logical_and(np.greater(dta[0], valid_min),
                                          np.less(dta[0], valid_max)),
                           (data - offset) / gain, nodata).astype('B')
        dt = InfoObject()
        dt.data = dt_data[::-1]
        dt.info = self.get_data_header(dt.info, dta)
        dt.info["nodata"] = nodata
        dt.info["gain"] = gain
        dt.info["offset"] = offset
        self.dt = dt
        self._projectables.append('dt')

        # Bias:
        bias = self.file.variables['sses_bias']
        gain = 0.01
        offset = -1.27
        nodata = 255
        x = bias[0] * bias.scale_factor + bias.add_offset
        valid_min = bias.valid_min
        valid_max = bias.valid_max
        bias_data = np.where(np.logical_and(np.greater(bias[0], valid_min),
                                            np.less(bias[0], valid_max)),
                             (x - offset) / gain, nodata).astype('B')

        bias_obj = InfoObject()
        bias_obj.data = bias_data[::-1]
        bias_obj.info = self.get_data_header(bias_obj.info, bias)
        bias_obj.info["nodata"] = nodata
        bias_obj.info["gain"] = gain
        bias_obj.info["offset"] = offset
        self.bias = bias_obj
        self._projectables.append('bias')

        # Standard deviation:
        stdv = self.file.variables['sses_standard_deviation']
        gain = 0.01
        offset = 0.0
        nodata = 255
        x = stdv[0] * stdv.scale_factor + stdv.add_offset
        valid_min = stdv.valid_min
        valid_max = stdv.valid_max
        stdv_data = np.where(np.logical_and(np.greater(stdv[0], valid_min),
                                            np.less(stdv[0], valid_max)),
                             (x - offset) / gain, nodata).astype('B')

        stdv_obj = InfoObject()
        stdv_obj.data = stdv_data[::-1]
        stdv_obj.info = self.get_data_header(stdv_obj.info, stdv)
        stdv_obj.info["nodata"] = nodata
        stdv_obj.info["gain"] = gain
        stdv_obj.info["offset"] = offset
        self.stdv = stdv_obj
        self._projectables.append('stdv')

        # L2P flags:
        l2pf = self.file.variables['l2p_flags'][0]
        l2pf_obj = InfoObject()
        l2pf_obj.data = l2pf[::-1]
        l2pf_obj.info = self.get_data_header(l2pf_obj.info, l2pf)
        self.l2pf = l2pf_obj
        self._projectables.append('l2pf')

        # Longitudes:
        lon = self.file.variables['lon']
        self.lon = InfoObject()
        self.lon.data = lon[::-1].astype('f')
        self.lon.info = self.get_data_header(self.lon.info, lon)

        # Latitudes:
        lat = self.file.variables['lat']
        self.lat = InfoObject()
        self.lat.data = lat[::-1].astype('f')
        self.lat.info = self.get_data_header(self.lat.info, lat)

        return

    def project(self, coverage):
        """Project the data"""
        LOG.debug("Projecting channel %s..." % (self.name))
        import copy
        res = copy.copy(self)

        # Project the data
        for var in self._projectables:
            LOG.info("Projecting " + str(var))
            res.__dict__[var] = copy.copy(self.__dict__[var])
            res.__dict__[var].data = coverage.project_array(
                self.__dict__[var].data)

        res.name = self.name
        res.resolution = self.resolution
        res.filled = True
        res.area = coverage.out_area
        return res

    def is_loaded(self):
        """Tells if the channel contains loaded data.
        """
        return True

    def get_data_header(self, header, datasetObj):
        """Get the attribute names in the datasets that are common"""
        import copy
        retv = copy.copy(header)
        try:
            retv["longt_name"] = datasetObj.long_name
        except AttributeError:
            print "WARNING: No attribute 'long_name'"
        try:
            retv["standard_name"] = datasetObj.standard_name
        except AttributeError:
            print "WARNING: No attribute 'standard_name'"
        try:
            retv["comment"] = datasetObj.comment
        except AttributeError:
            pass
        try:
            retv["units"] = datasetObj.units
        except AttributeError:
            pass
        try:
            retv["valid_min"] = datasetObj.valid_min
        except AttributeError:
            pass
        try:
            retv["valid_max"] = datasetObj.valid_max
        except AttributeError:
            pass
        try:
            retv["scale_factor"] = datasetObj.scale_factor
        except AttributeError:
            pass
        try:
            retv["add_offset"] = datasetObj.add_offset
        except AttributeError:
            pass

        return retv

    def fillheader(self):
        """Get global header info and store in object"""
        pltname = self.file.platform
        self.mda["platform"] = ''.join(pltname.split('_'))
        self.mda["sensor"] = self.file.sensor
        self.mda["spatial_resolution"] = self.file.spatial_resolution
        self.mda["title"] = self.file.title
        self.mda["comment"] = self.file.comment

        self.mda["contact"] = self.file.creator_email
        self.mda["institution"] = self.file.institution
        self.mda["start_time"] = datetime.strptime(
            self.file.start_time, '%Y%m%dT%H%M%SZ')
        self.mda["stop_time"] = datetime.strptime(
            self.file.stop_time, '%Y%m%dT%H%M%SZ')
        # Reference time of sst file in seconds since 1981-01-01 00:00:00:
        self.mda["time"] = self.file.variables['time'][0]
        self.mda['westernmost_longitude'] = self.file.westernmost_longitude
        self.mda['easternmost_longitude'] = self.file.easternmost_longitude
        self.mda['southernmost_latitude'] = self.file.southernmost_latitude
        self.mda['northernmost_latitude'] = self.file.northernmost_latitude


def get_filename(scene, area_name):

    conf = ConfigParser()
    conf.read(os.path.join(CONFIG_PATH, scene.fullname + ".cfg"))

    filename = conf.get(scene.instrument_name + "-level4",
                        "sst_product_filename",
                        raw=True,
                        vars=os.environ)
    directory = conf.get(scene.instrument_name + "-level4",
                         "sst_product_dir",
                         vars=os.environ)
    pathname_tmpl = os.path.join(directory, filename)
    LOG.debug("Path = " + str(pathname_tmpl))

    if not scene.orbit:
        orbit = ""
    else:
        orbit = scene.orbit

    filename_tmpl = (scene.time_slot.strftime(pathname_tmpl)
                     % {"area": area_name,
                        "satellite": scene.satname + scene.number})

    product = 'sst'
    file_list = glob.glob(filename_tmpl)
    if len(file_list) > 1:
        LOG.warning("More than 1 file matching for " + product + "! "
                    + str(file_list))
        return None
    elif len(file_list) == 0:
        LOG.warning(
            "No " + product + " matching!: " + filename_tmpl)
        return
    else:
        filename = file_list[0]

    return filename


class OSISAF_SST_Reader(Reader):

    pformat = "nc_osisaf_l2"

    def load(self, satscene, *args, **kwargs):
        """Read data from file and load it into *satscene*.
        """
        lonlat_is_loaded = False

        prodfilename = kwargs.get('filename')

        if "SST" not in satscene.channels_to_load:
            LOG.warning("No SST product requested. Nothing to be done...")
            return

        try:
            area_name = satscene.area_id or satscene.area.area_id
        except AttributeError:
            area_name = "satproj_?????_?????"

        conf = ConfigParser()
        conf.read(os.path.join(CONFIG_PATH, satscene.fullname + ".cfg"))

        # Reading the products
        product = "sst"
        classes = {product: OsisafNarSstProduct}

        LOG.debug("Loading " + product)

        if isinstance(prodfilename, (list, tuple, set)):
            for fname in prodfilename:
                kwargs['filename'] = fname
                self.load(satscene, *args, **kwargs)
            return
        elif (prodfilename and
              'OSISAF-L3C' in os.path.basename(prodfilename)):
            if os.path.basename(prodfilename).split("_")[2] == 'SST':
                filename = prodfilename
            else:
                LOG.warning(
                    "Product file name is not as expected: " + str(prodfilename))
                return
        else:
            filename = get_filename(satscene, area_name)

        LOG.info("Filename = " + str(filename))

        chn = classes[product]()
        chn.read(filename, lonlat_is_loaded == False)
        satscene.channels.append(chn)

        lons, lats = chn.lon.data, chn.lat.data
        lonlat_is_loaded = True

        nodata_mask = False
        try:
            from pyresample import geometry
            lons = np.ma.masked_array(lons, nodata_mask)
            lats = np.ma.masked_array(lats, nodata_mask)
            area = geometry.SwathDefinition(lons=lons,
                                            lats=lats)
        except ImportError:
            area = None

        if area:
            chn.area = area
        else:
            chn.lon = lons
            chn.lat = lats

        LOG.info("Loading OSISAF SST parameters done")

        return



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2011, 2012, 2013, 2014, 2015, 2016, 2017.

# Author(s):

#
#   Adam Dybbroe <adam.dybbroe@smhi.se>
#   Kristian Rune Larsen <krl@dmi.dk>
#   Lars Ørum Rasmussen <ras@dmi.dk>
#   Martin Raspaud <martin.raspaud@smhi.se>
#

# This file is part of mpop.

# mpop is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# mpop is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# mpop.  If not, see <http://www.gnu.org/licenses/>.

"""Interface to VIIRS SDR format

Format documentation:
http://npp.gsfc.nasa.gov/science/sciencedocuments/082012/474-00001-03_CDFCBVolIII_RevC.pdf

"""
import hashlib
import logging
import os.path
from ConfigParser import ConfigParser
from datetime import datetime, timedelta

import h5py
import numpy as np

from mpop import CONFIG_PATH
# ------------------------------------------------------------------------------
from mpop.plugin_base import Reader
from mpop.utils import strftime

NO_DATE = datetime(1958, 1, 1)
EPSILON_TIME = timedelta(days=2)
VIIRS_MBAND_GRANULE_SIZE = (768, 3200)
VIIRS_DNB_GRANULE_SIZE = (768, 4064)
VIIRS_IBAND_GRANULE_SIZE = (768 * 2, 3200 * 2)

VIIRS_IR_BANDS = ('M16', 'M15', 'M14', 'M13', 'M12', 'I5', 'I4')
VIIRS_VIS_BANDS = ('M1', 'M2', 'M3', 'M4', 'M5', 'M6',
                   'M7', 'M8', 'M9', 'M10', 'M11',
                   'I1', 'I2', 'I3')
VIIRS_DNB_BANDS = ('DNB', )

logger = logging.getLogger(__name__)


class HDF5MetaData(object):

    """

    Small class for inspecting a HDF5 file and retrieve its metadata/header
    data. It is developed for JPSS/NPP data but is really generic and should
    work on most other hdf5 files.

    Supports

    """

    def __init__(self, filename):
        self.metadata = {}
        self.filename = filename
        if not os.path.exists(filename):
            raise IOError("File %s does not exist!" % filename)

    def read(self):
        h5f = h5py.File(self.filename, 'r')
        h5f.visititems(self.collect_metadata)
        self._collect_attrs('/', h5f.attrs)
        h5f.close()
        return self

    def _collect_attrs(self, name, attrs):
        for key, value in attrs.iteritems():
            value = np.squeeze(value)
            if issubclass(value.dtype.type, str):
                self.metadata["%s/attr/%s" % (name, key)] = str(value)
            else:
                self.metadata["%s/attr/%s" % (name, key)] = value

    def collect_metadata(self, name, obj):
        if isinstance(obj, h5py.Dataset):
            self.metadata["%s/shape" % name] = obj.shape
        self._collect_attrs(name, obj.attrs)

    def __getitem__(self, key):

        long_key = None
        for mkey in self.metadata.keys():
            if mkey.endswith(key):
                if long_key is not None:
                    raise KeyError("Multiple keys called %s" % key)
                long_key = mkey
                # Test this to be able to read SDR files with several granules
                # in one file (Matias' files)
                break
        return self.metadata[long_key]

    def keys(self):
        return self.metadata.keys()

    def get_data_keys(self):

        data_keys = []
        for key in self.metadata.keys():
            if key.endswith("/shape"):
                data_key = key.split("/shape")[0]
                data_keys.append(data_key)
        return data_keys


class NPPMetaData(HDF5MetaData):

    def _parse_npp_datatime(self, datestr, timestr):
        time_val = datetime.strptime(datestr + timestr, '%Y%m%d%H%M%S.%fZ')
        if abs(time_val - NO_DATE) < EPSILON_TIME:
            raise ValueError("Datetime invalid %s " % time_val)
        return time_val

    def get_ring_lonlats(self):
        return self['G-Ring_Longitude'], self['G-Ring_Latitude']

    def get_begin_time(self):
        return self._parse_npp_datatime(self['AggregateBeginningDate'],
                                        self['AggregateBeginningTime'])

    def get_end_time(self):
        return self._parse_npp_datatime(self['AggregateEndingDate'],
                                        self['AggregateEndingTime'])

    def get_begin_orbit_number(self):
        return int(self['AggregateBeginningOrbitNumber'])

    def get_end_orbit_number(self):
        return int(self['AggregateEndingOrbitNumber'])

    def get_geofilname(self):
        return self['N_GEO_Ref']

    def get_shape(self):

        shape = self['Radiance/shape']
        band = self['Band_ID']
        if band[0] == 'M':
            # if shape != VIIRS_MBAND_GRANULE_SIZE:
            if ((shape[0] % VIIRS_MBAND_GRANULE_SIZE[0]) != 0 and
                    (shape[1] % VIIRS_MBAND_GRANULE_SIZE[1]) != 0):
                raise ValueError(
                    "Unsupported granule size %s for %s" % (shape, band))
        elif band == "DNB":
            if ((shape[0] % VIIRS_DNB_GRANULE_SIZE[0]) != 0 and
                    (shape[1] % VIIRS_MBAND_GRANULE_SIZE[1]) != 0):
                raise ValueError(
                    "Unsupported granule size %s for %s" % (shape, band))
        elif band[0] == "I":
            if ((shape[0] % VIIRS_IBAND_GRANULE_SIZE[0] != 0) and
                    (shape[1] % VIIRS_IBAND_GRANULE_SIZE[1] != 0)):
                raise ValueError(
                    "Unsupported granule size %s for %s" % (shape, band))

        return shape

    def get_band_description(self):

        band = self['Band_ID']

        for band_desc in ('I', 'M', 'DNB', "N/A"):
            if band.startswith(band_desc):
                if band_desc == 'N/A':
                    return 'DNB'
                return band_desc
        return None

    def _band_data_keys(self, data_type):
        """
        :param data_type: Reflectance, Radiance or BrightnessTemperature
        :type data_type: string
        :returns: HDF5 data key and scale factor keys in a two element tuple

        """
        data_key = None
        factors_keys = None
        for key in self.get_data_keys():
            if key.endswith(data_type):
                data_key = key
                factors_keys = key + "Factors"

        return (data_key, factors_keys)

    def get_reflectance_keys(self):
        return self._band_data_keys("Reflectance")

    def get_radiance_keys(self):
        return self._band_data_keys("Radiance")

    def get_brightness_temperature_keys(self):
        return self._band_data_keys("BrightnessTemperature")

    def get_unit(self, calibrate=1):

        band = self['Band_ID']
        if calibrate == 2 and band not in VIIRS_DNB_BANDS:
            return "W m-2 um-1 sr-1"

        if band in VIIRS_IR_BANDS:
            return "K"
        elif band in VIIRS_VIS_BANDS:
            return '%'
        elif band in VIIRS_DNB_BANDS:
            return 'W m-2 sr-1'

        return None


#
#
# http://yloiseau.net/articles/DesignPatterns/flyweight/
class GeolocationFlyweight(object):

    def __init__(self, cls):
        self._cls = cls
        self._instances = dict()

    def __call__(self, *args, **kargs):
        """
        we assume that this is only used for the gelocation object,
        filenames are listed in the second argument

        """
        return self._instances.setdefault(tuple(args[1]),
                                          self._cls(*args, **kargs))

    def clear_cache(self):
        del self._instances
        self._instances = dict()


@GeolocationFlyweight
class ViirsGeolocationData(object):

    def __init__(self, shape, filenames):
        self.filenames = filenames
        self.longitudes = None
        self.shape = shape
        self.latitudes = None
        self.elevation = None
        self.mask = None

    def read(self):
        """
        Read longitudes and latitudes from geo filenames and assemble
        """

        if self.longitudes is not None:
            return self

        self.longitudes = np.empty(self.shape,
                                   dtype=np.float32)
        self.latitudes = np.empty(self.shape,
                                  dtype=np.float32)
        self.elevation = np.empty(self.shape,
                                  dtype=np.float32)
        self.mask = np.zeros(self.shape,
                             dtype=np.bool)

        granule_length = self.shape[0] / len(self.filenames)

        for index, filename in enumerate(self.filenames):

            swath_index = index * granule_length
            y0_ = swath_index
            y1_ = swath_index + granule_length

            get_lonlat_into(filename,
                            self.longitudes[y0_:y1_, :],
                            self.latitudes[y0_:y1_, :],
                            self.elevation[y0_:y1_, :],
                            self.mask[y0_:y1_, :])

        self.longitudes = np.ma.array(self.longitudes,
                                      mask=self.mask,
                                      copy=False)
        self.latitudes = np.ma.array(self.latitudes,
                                     mask=self.mask,
                                     copy=False)
        self.elevation = np.ma.array(self.elevation,
                                     mask=self.mask,
                                     copy=False)
        logger.debug("Geolocation read in for... " + str(self))
        return self


# ------------------------------------------------------------------------------

def _get_invalid_info(granule_data):
    """Get a detailed report of the missing data.
        N/A: not applicable
        MISS: required value missing at time of processing
        OBPT: onboard pixel trim (overlapping/bow-tie pixel removed during
            SDR processing)
        OGPT: on-ground pixel trim (overlapping/bow-tie pixel removed
            during EDR processing)
        ERR: error occurred during processing / non-convergence
        ELINT: ellipsoid intersect failed / instrument line-of-sight does
            not intersect the Earth’s surface
        VDNE: value does not exist / processing algorithm did not execute
        SOUB: scaled out-of-bounds / solution not within allowed range
    """
    if issubclass(granule_data.dtype.type, np.integer):
        msg = ("na:" + str((granule_data == 65535).sum()) +
               " miss:" + str((granule_data == 65534).sum()) +
               " obpt:" + str((granule_data == 65533).sum()) +
               " ogpt:" + str((granule_data == 65532).sum()) +
               " err:" + str((granule_data == 65531).sum()) +
               " elint:" + str((granule_data == 65530).sum()) +
               " vdne:" + str((granule_data == 65529).sum()) +
               " soub:" + str((granule_data == 65528).sum()))
    if issubclass(granule_data.dtype.type, np.floating):
        msg = ("na:" + str((granule_data == -999.9).sum()) +
               " miss:" + str((granule_data == -999.8).sum()) +
               " obpt:" + str((granule_data == -999.7).sum()) +
               " ogpt:" + str((granule_data == -999.6).sum()) +
               " err:" + str((granule_data == -999.5).sum()) +
               " elint:" + str((granule_data == -999.4).sum()) +
               " vdne:" + str((granule_data == -999.3).sum()) +
               " soub:" + str((granule_data == -999.2).sum()))
    return msg


class ViirsBandData(object):

    """Placeholder for the VIIRS M&I-band data.
    Reads the SDR data - one hdf5 file for each band.
    Not yet considering the Day-Night Band
    """

    def __init__(self, filenames, calibrate=1):
        self.begin_time = 0
        self.end_time = 0
        self.orbit_begin = 0
        self.orbit_end = 0
        self.band_id = 'unknown'
        self.data = None
        self.mask = None
        self.raw_data = None
        self.scale = 1.0    # gain
        self.offset = 0.0   # intercept
        self.filenames = sorted(filenames)
        self.units = 'unknown'
        self.geo_filenames = []
        self.calibrate = calibrate

        self.data = None
        self.geolocation = None

        self.band_desc = None
        self.band_uid = None
        self.metadata = []

    def read(self):
        self._read_metadata()

        logger.debug("Shape of data: " + str(self.raw_data.shape))

        self._read_data()

        return self

    def _read_metadata(self):

        logger.debug("Filenames: " + str(self.filenames))

        for fname in self.filenames:
            logger.debug("Get and append metadata from file: " + str(fname))
            md = NPPMetaData(fname).read()
            self.metadata.append(md)
            self.geo_filenames.append(md.get_geofilname())
            # Check if the geo-filenames actually exists:
            # geofilename = md.get_geofilname()
            # if os.path.exists(geofilename):
            #     self.geo_filenames.append(geofilename)
            # else:
            #     logger.warning("Geo file defined in metadata header " +
            #                    "does not exist: " + str(geofilename))

        #
        # initiate data arrays
        granule_length, swath_width = self.metadata[0].get_shape()
        shape = (granule_length * len(self.metadata), swath_width)

        #self.data = np.ma.array(np.zeros(shape, dtype=np.float32), fill_value=0)
        self.raw_data = np.zeros(shape, dtype=np.float32)
        self.mask = np.zeros(shape, dtype=np.bool)
        self.orbit_begin = self.metadata[0].get_begin_orbit_number()
        self.orbit_end = self.metadata[-1].get_end_orbit_number()
        self.begin_time = self.metadata[0].get_begin_time()
        self.end_time = self.metadata[-1].get_end_time()

        self.units = self.metadata[0].get_unit(self.calibrate)
        self.band_desc = self.metadata[0].get_band_description()

        self.band_id = self.metadata[0]['Band_ID']
        if self.band_id == "N/A":
            self.band_id = "DNB"

    def _read_data(self):
        """Read one VIIRS M- or I-band channel: Data and attributes (meta data)

        - *calibrate* set to 1 (default) returns reflectances for visual bands,
           tb for ir bands, and radiance for dnb.

        - *calibrate* set to 2 returns radiances.
        """

        granule_length, swath_width = self.metadata[0].get_shape()

        for index, md in enumerate(self.metadata):
            h5f = h5py.File(md.filename, 'r')

            # find appropiate band data to insert
            data_key = None
            factors_key = None
            if self.calibrate == 1:
                data_key, factors_key = md.get_reflectance_keys()
                if data_key is None:
                    data_key, factors_key = md.get_brightness_temperature_keys()
                    logger.debug("data and factors keys: " +
                                 str(data_key) + str(factors_key))
                # handle dnb data
                if data_key is None and self.band_id == "DNB":
                    data_key, factors_key = md.get_radiance_keys()

            elif self.calibrate == 2:
                data_key, factors_key = md.get_radiance_keys()

            #
            # get granule data and scale and offset values
            try:
                granule_factors_data = h5f[factors_key].value
            except KeyError:
                #
                # We can't find the factors this must be DNB
                # if self.band_id != "DNB":
                #     raise
                # We can't find the factors, so this must be the DNB or the M13
                # band:
                logger.debug("Band id = " + str(self.band_id))

                if self.band_id != "DNB" and self.band_id != "M13":
                    raise
                if self.band_id == "DNB":
                    # The unit is W/sr cm-2 in the file! but we need 'W sr-1
                    # m-2'
                    granule_factors_data = (10000., 0.)
                else:
                    # M13 is stored in floats with no scale or offset:
                    granule_factors_data = (1., 0.)

            granule_data = h5f[data_key].value

            self.scale, self.offset = granule_factors_data[0:2]

            # The VIIRS reflectances are between 0 and 1.
            # mpop standard is '%'
            if self.units == '%':
                # To get reflectances in percent!
                self.scale *= np.int8(100)
                self.offset *= np.int8(100)

            swath_index = index * granule_length
            y0_ = swath_index
            y1_ = swath_index + granule_length

            self.raw_data[y0_:y1_, :] = granule_data
            self.raw_data[y0_:y1_, :] *= self.scale
            self.raw_data[y0_:y1_, :] += self.offset
            logger.debug("dtype(granule_data) = " + str(granule_data.dtype))

            # Masking spurious data
            # according to documentation, mask integers >= 65328, floats <=
            # -999.3

            if issubclass(granule_data.dtype.type, np.integer):
                self.mask[y0_:y1_, :] = granule_data >= 65528
            if issubclass(granule_data.dtype.type, np.floating):
                self.mask[y0_:y1_, :] = granule_data < -999.2
            logger.debug("%s", _get_invalid_info(granule_data))

            # Is it necessary to mask negatives?
            #self.mask[y0_:y1_, :] |= self.raw_data[y0_:y1_, :] < 0

            h5f.close()

        self.data = np.ma.array(self.raw_data, mask=self.mask, copy=False)

        self.band_uid = self.band_desc + hashlib.sha1(self.mask).hexdigest()

    def read_lonlat(self, geofilepaths=None, geodir=None):

        if geofilepaths is None:
            if geodir is None:
                geodir = os.path.dirname(self.metadata[0].filename)

            logger.debug("Geo-files = " + str(self.geo_filenames))
            geofilepaths = [os.path.join(geodir, geofilepath)
                            for geofilepath in self.geo_filenames]

        geofilepaths = sorted(geofilepaths)
        logger.debug("Geo-files = " + str(geofilepaths))
        self.geolocation = ViirsGeolocationData(self.data.shape,
                                                geofilepaths).read()


class ViirsSDRReader(Reader):
    pformat = "viirs_sdr"

    def __init__(self, *args, **kwargs):
        Reader.__init__(self, *args, **kwargs)
        self.geofiles = []
        self.shape = None

    def get_sunsat_angles(self, **kwargs):
        """Get sun-satellite viewing geometry for a given band type (M, I, or
        DNB)
        Optional arguments:
            bandtype = 'M', 'I', or 'DNB'
        Return
            sun-zenith, sun-azimuth, sat-zenith, sat-azimuth

        """

        if 'bandtype' in kwargs:
            bandtype = kwargs['bandtype']
        else:
            bandtype = 'M'

        if bandtype.startswith('M'):
            geofilenames = [geofile for geofile in self.geofiles
                            if os.path.basename(geofile).startswith('GMTCO')]
            if len(geofilenames) == 0:
                # Try the geoid instead:
                geofilenames = [geofile for geofile in self.geofiles
                                if os.path.basename(geofile).startswith('GMODO')]
        elif bandtype.startswith('I'):
            geofilenames = [geofile for geofile in self.geofiles
                            if os.path.basename(geofile).startswith('GITCO')]
            if len(geofilenames) == 0:
                # Try the geoid instead:
                geofilenames = [geofile for geofile in self.geofiles
                                if os.path.basename(geofile).startswith('GIMGO')]
        elif bandtype.startswith('DNB'):
            geofilenames = [geofile for geofile in self.geofiles
                            if os.path.basename(geofile).startswith('GDNBO')]

        else:
            logger.error("Band type %s not supported", bandtype)
            return None

        geofilenames = sorted(geofilenames)

        data = {}
        mask = {}
        h5names = ['SolarZenithAngle', 'SolarAzimuthAngle',
                   'SatelliteZenithAngle', 'SatelliteAzimuthAngle']
        local_names = ['sunz', 'sun_azi',
                       'satz', 'sat_azi']
        for item in local_names:
            data[item] = np.empty(self.shape,
                                  dtype=np.float32)
            mask[item] = np.zeros(self.shape,
                                  dtype=np.bool)

        granule_length = self.shape[0] / len(geofilenames)

        for index, filename in enumerate(geofilenames):

            swath_index = index * granule_length
            y0_ = swath_index
            y1_ = swath_index + granule_length

            for angle, param_name in zip(h5names, local_names):
                get_viewing_angle_into(filename,
                                       data[param_name][y0_:y1_, :],
                                       mask[param_name][y0_:y1_, :], angle)

        for item in local_names:
            data[item] = np.ma.array(data[item], mask=mask[item], copy=False)

        return (data['sunz'], data['sun_azi'],
                data['satz'], data['sat_azi'])

    def get_elevation(self, **kwargs):
        """Get elevation/topography for a given band type (M, I, or
        DNB)
        Optional arguments:
            bandtype = 'M', 'I', or 'DNB'
        Return
            elevation

        """

        if 'bandtype' in kwargs:
            bandtype = kwargs['bandtype']
        else:
            bandtype = 'M'

        if bandtype.startswith('M'):
            geofilenames = [geofile for geofile in self.geofiles
                            if os.path.basename(geofile).startswith('GMTCO')]
            if len(geofilenames) == 0:
                # Try the geoid instead:
                geofilenames = [geofile for geofile in self.geofiles
                                if os.path.basename(geofile).startswith('GMODO')]
        elif bandtype.startswith('I'):
            geofilenames = [geofile for geofile in self.geofiles
                            if os.path.basename(geofile).startswith('GITCO')]
            if len(geofilenames) == 0:
                # Try the geoid instead:
                geofilenames = [geofile for geofile in self.geofiles
                                if os.path.basename(geofile).startswith('GIMGO')]
        elif bandtype.startswith('DNB'):
            geofilenames = [geofile for geofile in self.geofiles
                            if os.path.basename(geofile).startswith('GDNBO')]

        else:
            logger.error("Band type %s not supported", bandtype)
            return None

        geofilenames = sorted(geofilenames)

        hdata = np.empty(self.shape,
                         dtype=np.float32)
        hmask = np.zeros(self.shape,
                         dtype=np.bool)

        granule_length = self.shape[0] / len(geofilenames)

        for index, filename in enumerate(geofilenames):

            swath_index = index * granule_length
            y0_ = swath_index
            y1_ = swath_index + granule_length

            get_elevation_into(filename,
                               hdata[y0_:y1_, :],
                               hmask[y0_:y1_, :])

        hdata = np.ma.array(hdata, mask=hmask, copy=False)

        return hdata

    def load(self, satscene, calibrate=1, time_interval=None,
             area=None, filename=None, **kwargs):
        """Read viirs SDR reflectances and Tbs from file and load it into
        *satscene*.
        """
        if satscene.instrument_name != "viirs":
            raise ValueError("Wrong instrument, expecting viirs")

        if kwargs:
            logger.warning(
                "Unsupported options for viirs reader: %s", str(kwargs))

        conf = ConfigParser()
        conf.read(os.path.join(CONFIG_PATH, satscene.fullname + ".cfg"))
        options = {}
        for option, value in conf.items(satscene.instrument_name + "-level2",
                                        raw=True):
            options[option] = value

        band_list = [s.name for s in satscene.channels]
        chns = satscene.channels_to_load & set(band_list)
        if len(chns) == 0:
            return

        if time_interval:
            time_start, time_end = time_interval
        else:
            time_start, time_end = satscene.time_slot, None

        import glob

        if "filename" not in options:
            raise IOError("No filename given, cannot load")

        values = {"orbit": satscene.orbit,
                  "satname": satscene.satname,
                  "instrument": satscene.instrument_name,
                  "satellite": satscene.satname
                  #"satellite": satscene.fullname
                  }

        file_list = []
        if filename is not None:
            if not isinstance(filename, (list, set, tuple)):
                filename = [filename]
            geofile_list = []
            for fname in filename:
                if os.path.basename(fname).startswith("SV"):
                    file_list.append(fname)
                elif os.path.basename(fname).startswith("G"):
                    geofile_list.append(fname)
                else:
                    logger.info("Unrecognized SDR file: %s", fname)
            if file_list:
                directory = os.path.dirname(file_list[0])
            if geofile_list:
                geodirectory = os.path.dirname(geofile_list[0])

        if not file_list:
            filename_tmpl = strftime(
                satscene.time_slot, options["filename"]) % values

            directory = strftime(satscene.time_slot, options["dir"]) % values

            if not os.path.exists(directory):
                #directory = globify(options["dir"]) % values
                directory = globify(
                    strftime(satscene.time_slot, options["dir"])) % values
                logger.debug(
                    "Looking for files in directory " + str(directory))
                directories = glob.glob(directory)
                if len(directories) > 1:
                    raise IOError("More than one directory for npp scene... " +
                                  "\nSearch path = %s\n\tPlease check npp.cfg file!" % directory)
                elif len(directories) == 0:
                    raise IOError("No directory found for npp scene. " +
                                  "\nSearch path = %s\n\tPlease check npp.cfg file!" % directory)
                else:
                    directory = directories[0]

            file_list = glob.glob(os.path.join(directory, filename_tmpl))

            # Only take the files in the interval given:
            logger.debug("Number of files before segment selection: "
                         + str(len(file_list)))
            for fname in file_list:
                if os.path.basename(fname).startswith("SVM14"):
                    logger.debug("File before segmenting: "
                                 + os.path.basename(fname))
            file_list = _get_swathsegment(
                file_list, time_start, time_end, area)
            logger.debug("Number of files after segment selection: "
                         + str(len(file_list)))

            for fname in file_list:
                if os.path.basename(fname).startswith("SVM14"):
                    logger.debug("File after segmenting: "
                                 + os.path.basename(fname))

            logger.debug("Template = " + str(filename_tmpl))

            # 22 VIIRS bands (16 M-bands + 5 I-bands + DNB)
            if len(file_list) % 22 != 0:
                logger.warning("Number of SDR files is not divisible by 22!")
            if len(file_list) == 0:
                logger.debug(
                    "File template = " + str(os.path.join(directory, filename_tmpl)))
                raise IOError("No VIIRS SDR file matching!: " +
                              "Start time = " + str(time_start) +
                              "  End time = " + str(time_end))

            geo_dir_string = options.get("geo_dir", None)
            if geo_dir_string:
                geodirectory = strftime(
                    satscene.time_slot, geo_dir_string) % values
            else:
                geodirectory = directory
            logger.debug("Geodir = " + str(geodirectory))

            geofile_list = []
            geo_filenames_string = options.get("geo_filenames", None)
            if geo_filenames_string:
                geo_filenames_tmpl = strftime(satscene.time_slot,
                                              geo_filenames_string) % values
                geofile_list = glob.glob(os.path.join(geodirectory,
                                                      geo_filenames_tmpl))
                logger.debug("List of geo-files: " + str(geofile_list))
                # Only take the files in the interval given:
                geofile_list = _get_swathsegment(
                    geofile_list, time_start, time_end)

            logger.debug("List of geo-files (after time interval selection): "
                         + str(geofile_list))

        filenames = [os.path.basename(s) for s in file_list]

        glob_info = {}

        self.geofiles = geofile_list

        logger.debug("Channels to load: " + str(satscene.channels_to_load))
        for chn in satscene.channels_to_load:
            # Take only those files in the list matching the band:
            # (Filename starts with 'SV' and then the band-name)
            fnames_band = []

            try:
                fnames_band = [s for s in filenames if s.find('SV' + chn) >= 0]
            except TypeError:
                logger.warning('Band frequency not available from VIIRS!')
                logger.info('Asking for channel' + str(chn) + '!')

            if len(fnames_band) == 0:
                continue

            filename_band = [
                os.path.join(directory, fname) for fname in fnames_band]
            logger.debug("fnames_band = " + str(filename_band))

            band = ViirsBandData(filename_band, calibrate=calibrate).read()
            logger.debug('Band id = ' + band.band_id)

            # If the list of geo-files is not specified in the config file or
            # some of them do not exist, we rely on what is written in the
            # band-data metadata header:
            if len(geofile_list) < len(filename_band):
                geofilenames_band = [os.path.join(geodirectory, gfile) for
                                     gfile in band.geo_filenames]
                logger.debug("Geolocation filenames for band: " +
                             str(geofilenames_band))
                # Check if the geo-filenames found from the metadata actually
                # exist and issue a warning if they do not:
                for geofilename in geofilenames_band:
                    if not os.path.exists(geofilename):
                        logger.warning("Geo file defined in metadata header " +
                                       "does not exist: " + str(geofilename))

            elif band.band_id.startswith('M'):
                geofilenames_band = [geofile for geofile in geofile_list
                                     if os.path.basename(geofile).startswith('GMTCO')]
                if len(geofilenames_band) != len(filename_band):
                    # Try the geoid instead:
                    num_geo = len(geofilenames_band)
                    geofilenames_band = [geofile for geofile in geofile_list
                                         if os.path.basename(geofile).startswith('GMODO')]
                    if len(geofilenames_band) != len(filename_band):
                        num_geo = max(num_geo, len(geofilenames_band))
                        num_band = len(filename_band)
                        raise IOError("Incomplete dataset. " +
                                      "Found %d geolocation files" % num_geo +
                                      " and %d %s files" % (num_band,
                                                            band.band_id))
            elif band.band_id.startswith('I'):
                geofilenames_band = [geofile for geofile in geofile_list
                                     if os.path.basename(geofile).startswith('GITCO')]
                if len(geofilenames_band) != len(filename_band):
                    # Try the geoid instead:
                    num_geo = len(geofilenames_band)
                    geofilenames_band = [geofile for geofile in geofile_list
                                         if os.path.basename(geofile).startswith('GIMGO')]
                    if len(geofilenames_band) != len(filename_band):
                        num_geo = max(num_geo, len(geofilenames_band))
                        num_band = len(filename_band)
                        raise IOError("Incomplete dataset. " +
                                      "Found %d geolocation files" % num_geo +
                                      " and %d %s files" % (num_band,
                                                            band.band_id))
            elif band.band_id.startswith('D'):
                geofilenames_band = [geofile for geofile in geofile_list
                                     if os.path.basename(geofile).startswith('GDNBO')]
                if len(geofilenames_band) != len(filename_band):
                    num_geo = len(geofilenames_band)
                    num_band = len(filename_band)
                    raise IOError("Incomplete dataset. " +
                                  "Found %d GDNBO files " % num_geo +
                                  "and %d %s files" % (num_band, band.band_id))

            band.read_lonlat(geofilepaths=geofilenames_band)

            if not band.band_desc:
                logger.warning('Band name = ' + band.band_id)
                raise AttributeError('Band description not supported!')

            satscene[chn].data = band.data
            satscene[chn].info['units'] = band.units
            satscene[chn].info['band_id'] = band.band_id
            satscene[chn].info['start_time'] = band.begin_time
            satscene[chn].info['end_time'] = band.end_time
            if chn in ['M01', 'M02', 'M03', 'M04', 'M05', 'M06', 'M07',
                       'M08', 'M09', 'M10', 'M11', 'I01', 'I02', 'I03']:
                satscene[chn].info['sun_zen_correction_applied'] = True

            # We assume the same geolocation should apply to all M-bands!
            # ...and the same to all I-bands:

            from pyresample import geometry

            satscene[chn].area = geometry.SwathDefinition(
                lons=np.ma.masked_where(band.data.mask,
                                        band.geolocation.longitudes,
                                        copy=True),
                lats=np.ma.masked_where(band.data.mask,
                                        band.geolocation.latitudes,
                                        copy=True))
            height = np.ma.masked_where(band.data.mask,
                                        band.geolocation.elevation,
                                        copy=True)

            area_name = ("swath_" + satscene.fullname + "_" +
                         str(satscene.time_slot) + "_"
                         + str(satscene[chn].data.shape) + "_" +
                         band.band_uid)

            satscene[chn].area.area_id = area_name
            satscene[chn].area_id = area_name

            if self.shape is None:
                self.shape = band.data.shape

            # except ImportError:
            #    satscene[chn].area = None
            #    satscene[chn].lat = np.ma.array(band.latitude, mask=band.data.mask)
            #    satscene[chn].lon = np.ma.array(band.longitude, mask=band.data.mask)

            # if 'institution' not in glob_info:
            ##    glob_info['institution'] = band.global_info['N_Dataset_Source']
            # if 'mission_name' not in glob_info:
            ##    glob_info['mission_name'] = band.global_info['Mission_Name']

        ViirsGeolocationData.clear_cache()

        # Compulsory global attribudes
        satscene.info["title"] = (satscene.satname.capitalize() +
                                  " satellite, " +
                                  satscene.instrument_name.capitalize() +
                                  " instrument.")
        if 'institution' in glob_info:
            satscene.info["institution"] = glob_info['institution']

        if 'mission_name' in glob_info:
            satscene.add_to_history(glob_info['mission_name'] +
                                    " VIIRS SDR read by mpop")
        else:
            satscene.add_to_history("NPP/JPSS VIIRS SDR read by mpop")

        satscene.info["references"] = "No reference."
        satscene.info["comments"] = "No comment."

        satscene.info["start_time"] = min([chn.info["start_time"]
                                           for chn in satscene
                                           if chn.is_loaded()])
        satscene.info["end_time"] = max([chn.info["end_time"]
                                         for chn in satscene
                                         if chn.is_loaded()])


def get_lonlat_into(filename, out_lons, out_lats, out_height, out_mask):
    """Read lon,lat from hdf5 file"""
    logger.debug("Geo File = " + filename)

    md = HDF5MetaData(filename).read()

    h5f = h5py.File(filename, 'r')
    for key in md.get_data_keys():
        if key.endswith("Latitude"):
            h5f[key].read_direct(out_lats)
            out_mask[:] = out_lats < -999
        if key.endswith("Longitude"):
            h5f[key].read_direct(out_lons)
        if key.endswith("Height"):
            h5f[key].read_direct(out_height)
    h5f.close()


def get_elevation_into(filename, out_height, out_mask):
    """Read elevation/height from hdf5 file"""
    logger.debug("Geo File = " + filename)

    md = HDF5MetaData(filename).read()

    h5f = h5py.File(filename, 'r')
    for key in md.get_data_keys():
        if key.endswith("Height"):
            h5f[key].read_direct(out_height)
            out_mask[:] = out_height < -999
    h5f.close()


def get_viewing_angle_into(filename, out_val, out_mask, param):
    """Read a sun-sat viewing angle from hdf5 file"""
    logger.debug("Sun-Sat viewing geometry = " + filename)

    if param not in ['SolarZenithAngle',
                     'SolarAzimuthAngle',
                     'SatelliteZenithAngle',
                     'SatelliteAzimuthAngle']:
        logger.warning('Viewing geometry parameter %s not supported!', param)
        return None

    md = HDF5MetaData(filename).read()

    h5f = h5py.File(filename, 'r')
    for key in md.get_data_keys():
        if key.endswith('/' + param):
            h5f[key].read_direct(out_val)
            out_mask[:] = out_val < -999
    h5f.close()


def globify(filename):
    filename = filename.replace("%Y", "????")
    filename = filename.replace("%m", "??")
    filename = filename.replace("%d", "??")
    filename = filename.replace("%H", "??")
    filename = filename.replace("%M", "??")
    filename = filename.replace("%S", "??")
    return filename


def _get_times_from_npp(filename):

    bname = os.path.basename(filename)
    sll = bname.split('_')
    start_time = datetime.strptime(sll[2] + sll[3][:-1],
                                   "d%Y%m%dt%H%M%S")
    end_time = datetime.strptime(sll[2] + sll[4][:-1],
                                 "d%Y%m%de%H%M%S")
    if end_time < start_time:
        end_time += timedelta(days=1)
    return start_time, end_time


def _get_swathsegment(filelist, time_start, time_end=None, area=None):
    """
    Return only the granule files for the time interval or area.
    """
    if area is not None:
        from trollsched.spherical import SphPolygon
        from trollsched.boundary import AreaBoundary

        lons, lats = area.get_boundary_lonlats()
        area_boundary = AreaBoundary((lons.side1, lats.side1),
                                     (lons.side2, lats.side2),
                                     (lons.side3, lats.side3),
                                     (lons.side4, lats.side4))
        area_boundary.decimate(500)
        contour_poly = area_boundary.contour_poly

    segment_files = []
    for filename in filelist:

        timetup = _get_times_from_npp(filename)

        # Search for multiple granules using an area
        if area is not None:
            md = NPPMetaData(filename)
            md.read()
            coords = np.vstack(md.get_ring_lonlats())
            poly = SphPolygon(np.deg2rad(coords))
            if poly.intersection(contour_poly) is not None:
                segment_files.append(filename)
            continue

        # Search for single granule using time start
        if time_end is None:
            if time_start >= timetup[0] and time_start <= timetup[1]:
                segment_files.append(filename)
                continue

        # search for multiple granules
        else:
            # check that granule start time is inside interval
            if timetup[0] >= time_start and timetup[0] <= time_end:
                segment_files.append(filename)
                continue

            # check that granule end time is inside interval
            if timetup[1] >= time_start and timetup[1] <= time_end:
                segment_files.append(filename)
                continue

    segment_files.sort()
    return segment_files



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2009, 2010, 2011, 2012, 2013, 2014.

# SMHI,
# Folkborgsvägen 1,
# Norrköping,
# Sweden

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>
#   Adam Dybbroe <adam.dybbroe@smhi.se>

# This file is part of mpop.

# mpop is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# mpop is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with mpop.  If not, see <http://www.gnu.org/licenses/>.

"""This module handles coverage objects. Such objects are used to
transform area projected data by changing either the area or the
projection or both. A typical usage is to transform one large area in
satellite projection to an area of interest in polar projection for
example.
"""
import os
import ConfigParser
import logging

import numpy as np
from pyresample import image, utils, geometry, kd_tree
from pyresample.bilinear import get_sample_from_bil_info, get_bil_info
try:
    from pyresample.ewa import ll2cr, fornav
except ImportError:
    ll2cr, fornav = None, None

from mpop import CONFIG_PATH

logger = logging.getLogger(__name__)

area_file = None


def get_area_file():
    global area_file
    if area_file:
        return area_file

    conf = ConfigParser.ConfigParser()
    conf.read(os.path.join(CONFIG_PATH, "mpop.cfg"))

    try:
        area_file = os.path.join(conf.get("projector",
                                          "area_directory") or
                                 CONFIG_PATH,
                                 conf.get("projector", "area_file"))
    except ConfigParser.NoSectionError:
        area_file = ""
        logger.warning("Couldn't find the mpop.cfg file. "
                       "Do you have one ? is it in $PPP_CONFIG_DIR ?")
    return area_file


def get_area_def(area_name):
    """Get the definition of *area_name* from file. The file is defined to use
    is to be placed in the $PPP_CONFIG_DIR directory, and its name is defined
    in mpop's configuration file.
    """
    return utils.parse_area_file(get_area_file(), area_name)[0]


def _get_area_hash(area):
    """Calculate a (close to) unique hash value for a given area.
    """
    try:
        return hash(area.lons.tostring() + area.lats.tostring())
    except AttributeError:
        try:
            return hash(area.tostring())
        except AttributeError:
            return hash(str(area))


class Projector(object):

    """This class define projector objects. They contain the mapping
    information necessary for projection purposes. For efficiency reasons,
    generated projectors can be saved to disk for later reuse. Use the
    :meth:`save` method for this.

    To define a projector object, on has to specify *in_area* and
    *out_area*, and can also input the *in_lonlats* or the *mode*.
    Available modes area:
    - 'quick' (works only if both in- and out-areas are AreaDefinitions)
    - 'bilinear' (out-area needs to be AreaDefinition with proj4_string)
    - 'ewa'
    - 'nearest'.
    *radius* defines the radius of influence for nearest neighbour
    search in 'nearest' and 'bilinear' modes.
    """

    def __init__(self, in_area, out_area,
                 in_latlons=None, mode=None,
                 radius=10000, nprocs=1):

        if (mode is not None and
                mode not in ["quick", "nearest", "ewa", "bilinear"]):
            raise ValueError("Projector mode must be one of 'nearest', "
                             "'quick', 'ewa', 'bilinear'")

        self.area_file = get_area_file()

        self.in_area = None
        self.out_area = None
        self._cache = None
        self._filename = None
        self.mode = "quick"
        self.radius = radius
        self.conf = ConfigParser.ConfigParser()
        self.conf.read(os.path.join(CONFIG_PATH, "mpop.cfg"))

        # TODO:
        # - Rework so that in_area and out_area can be lonlats.
        # - Add a recompute flag ?

        # Setting up the input area
        in_id = self._setup_input_area(in_area, latlons=in_latlons)

        # Setting up the output area
        out_id = self._setup_output_area(out_area, latlons=None)

        # if self.in_area == self.out_area:
        #    return

        # choosing the right mode if necessary
        if mode is None:
            try:
                dicts = in_area.proj_dict, out_area.proj_dict
                del dicts
                self.mode = "quick"
            except AttributeError:
                self.mode = "nearest"
        else:
            self.mode = mode

        projections_directory = "/var/tmp"
        try:
            projections_directory = self.conf.get("projector",
                                                  "projections_directory")
        except ConfigParser.NoSectionError:
            pass

        self._filename = get_precompute_cache_fname(in_id, out_id,
                                                    in_area, out_area,
                                                    self.mode,
                                                    projections_directory)

        try:
            self._cache = {}
            self._file_cache = np.load(self._filename)
        except:
            logger.info("Computing projection from %s to %s...",
                        in_id, out_id)

            if self.mode == "nearest":
                self._cache = calc_nearest_params(in_area, out_area,
                                                  radius, nprocs=nprocs)

            elif self.mode == "quick":
                self._cache = calc_quick_params(in_area, out_area)

            elif self.mode == "ewa":
                if ll2cr is not None:
                    self._cache = calc_ewa_params(in_area, out_area)
                else:
                    raise ImportError("Can't import pyresample.ewa")

            elif self.mode == "bilinear":
                self._cache = calc_bilinear_params(in_area, out_area,
                                                   radius, nprocs=nprocs)

    def _setup_input_area(self, area, latlons=None):
        """Setup self.in_area and return area id"""
        try:
            self.in_area, in_id = get_area_and_id(area, latlons=latlons)
        except TypeError:
            raise utils.AreaNotFound("Input area " +
                                     str(area) +
                                     " must be defined in " +
                                     self.area_file +
                                     ", be an area object"
                                     " or longitudes/latitudes must be "
                                     "provided.")

        return in_id

    def _setup_output_area(self, area, latlons=None):
        """Setup output area"""
        try:
            self.out_area, out_id = get_area_and_id(area, latlons=latlons)
        except AttributeError:
            raise utils.AreaNotFound("Output area " +
                                     str(area) +
                                     " must be defined in " +
                                     self.area_file + " or "
                                     "be an area object.")
        return out_id

    def save(self, resave=False):
        """Save the precomputation to disk, and overwrite existing file in case
        *resave* is true.
        """
        if (not os.path.exists(self._filename)) or resave:
            logger.info("Saving projection to " +
                        self._filename)
            np.savez(self._filename, **self._cache)

    def _project_array_nearest(self, data):
        """Project array *data* using nearest neighbour resampling"""
        if 'valid_index' not in self._cache:
            self._cache['valid_index'] = self._file_cache['valid_index']
            self._cache['valid_output_index'] = \
                self._file_cache['valid_output_index']
            self._cache['index_array'] = self._file_cache['index_array']

        valid_index, valid_output_index, index_array = \
            (self._cache['valid_index'],
             self._cache['valid_output_index'],
             self._cache['index_array'])

        res = kd_tree.get_sample_from_neighbour_info('nn',
                                                     self.out_area.shape,
                                                     data,
                                                     valid_index,
                                                     valid_output_index,
                                                     index_array,
                                                     fill_value=None)
        return res

    def _project_array_quick(self, data):
        """Project array *data* using quick interpolation"""
        if 'row_idx' not in self._cache:
            self._cache['row_idx'] = self._file_cache['row_idx']
            self._cache['col_idx'] = self._file_cache['col_idx']
        row_idx, col_idx = self._cache['row_idx'], self._cache['col_idx']
        img = image.ImageContainer(data, self.in_area, fill_value=None)
        res = np.ma.array(img.get_array_from_linesample(row_idx, col_idx),
                          dtype=data.dtype)

        return res

    def _project_array_ewa(self, data):
        """Project array *data* using EWA interpolation"""
        # TODO: should be user configurable?
        rows_per_scan = None

        if 'ewa_cols' in self. not_cache:
            self._cache['ewa_cols'] = self._file_cache['ewa_cols']
            self._cache['ewa_rows'] = self._file_cache['ewa_rows']
        num_valid_points, res = fornav(self._cache['ewa_cols'],
                                       self._cache['ewa_rows'],
                                       self.out_area, data,
                                       rows_per_scan=rows_per_scan)
        del num_valid_points

        return res

    def _project_array_bilinear(self, data):
        """Project array *data* using bilinear interpolation"""
        if 'bilinear_t' not in self._cache:
            self._cache['bilinear_t'] = self._file_cache['bilinear_t']
            self._cache['bilinear_s'] = self._file_cache['bilinear_s']
            self._cache['input_idxs'] = self._file_cache['input_idxs']
            self._cache['idx_arr'] = self._file_cache['idx_arr']

        res = get_sample_from_bil_info(data.ravel(),
                                       self._cache['bilinear_t'],
                                       self._cache['bilinear_s'],
                                       self._cache['input_idxs'],
                                       self._cache['idx_arr'],
                                       output_shape=self.out_area.shape)
        res = np.ma.masked_invalid(res)

        return res

    def project_array(self, data):
        """Project an array *data* along the given Projector object.
        """

        if self.mode == "nearest":
            res = self._project_array_nearest(data)

        elif self.mode == "quick":
            res = self._project_array_quick(data)

        elif self.mode == "ewa":
            if fornav is not None:
                res = self._project_array_ewa(data)
            else:
                raise ImportError("Can't import pyresample.ewa")

        elif self.mode == "bilinear":
            res = self._project_array_bilinear(data)

        return res


def calc_nearest_params(in_area, out_area, radius, nprocs=1):
    """Calculate projection parameters for nearest neighbour
    interpolation"""
    valid_index, valid_output_index, index_array, distance_array = \
        kd_tree.get_neighbour_info(in_area,
                                   out_area,
                                   radius,
                                   neighbours=1,
                                   nprocs=nprocs)
    del distance_array
    cache = {}
    cache['valid_index'] = valid_index
    cache['valid_output_index'] = valid_output_index
    cache['index_array'] = index_array

    return cache


def calc_quick_params(in_area, out_area):
    """Calculate projection parameters for quick interpolation mode"""
    ridx, cidx = utils.generate_quick_linesample_arrays(in_area,
                                                        out_area)
    cache = {}
    cache['row_idx'] = ridx
    cache['col_idx'] = cidx

    return cache


def calc_bilinear_params(in_area, out_area, radius, nprocs=1):
    """Calculate projection parameters for bilinear interpolation"""
    bilinear_t, bilinear_s, input_idxs, idx_arr = \
        get_bil_info(in_area, out_area, radius, neighbours=32,
                     nprocs=nprocs, masked=False)

    cache = {}
    cache['bilinear_s'] = bilinear_s
    cache['bilinear_t'] = bilinear_t
    cache['input_idxs'] = input_idxs
    cache['idx_arr'] = idx_arr

    return cache


def calc_ewa_params(in_area, out_area):
    """Calculate projection parameters for EWA interpolation"""
    swath_points_in_grid, cols, rows = ll2cr(in_area, out_area)
    del swath_points_in_grid
    cache = {}
    # cache['ewa_swath_points_in_grid'] = \
    #     swath_points_in_grid
    cache['ewa_cols'] = cols
    cache['ewa_rows'] = rows

    return cache


def get_precompute_cache_fname(in_id, out_id, in_area, out_area, mode,
                               proj_dir):
    """Form filename for precompute cache"""

    filename = (in_id + "2" + out_id + "_" +
                str(_get_area_hash(in_area)) + "to" +
                str(_get_area_hash(out_area)) + "_" +
                mode + ".npz")

    return os.path.join(proj_dir, filename)


def get_area_and_id(area, latlons=None):
    try:
        area_def = get_area_def(area)
        area_id = area
    except (utils.AreaNotFound, AttributeError):
        try:
            area_id = area.area_id
            area_def = area
        except AttributeError:
            if latlons is None:
                raise
            else:
                # TODO: Note that latlons are in order (lons, lats)
                area_def = geometry.SwathDefinition(lons=latlons[0],
                                                    lats=latlons[1])
                area_id = area

    return area_def, area_id



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2010, 2011, 2012, 2014, 2016.

# Author(s):

#   Kristian Rune Larssen <krl@dmi.dk>
#   Adam Dybbroe <adam.dybbroe@smhi.se>
#   Martin Raspaud <martin.raspaud@smhi.se>
#   Esben S. Nielsen <esn@dmi.dk>

# This file is part of mpop.

# mpop is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# mpop is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# mpop.  If not, see <http://www.gnu.org/licenses/>.

"""The :mod:`satout.cfscene` module provide a proxy class and utilites for
conversion of mpop scene to cf conventions.
"""

import numpy as np
import numpy.ma as ma
from netCDF4 import date2num

from mpop.channel import Channel
from mpop.satout.netcdf4 import netcdf_cf_writer

import logging
LOG = logging.getLogger('cfscene')


#CF_DATA_TYPE = np.int16
CF_FLOAT_TYPE = np.float64
TIME_UNITS = "seconds since 1970-01-01 00:00:00"


class InfoObject(object):

    """Simple data and info container.
    """

    def __init__(self):
        self.info = {}
        self.data = None


class CFScene(object):

    """Scene proxy class for cf conventions. The constructor should be called
    with the *scene* to transform as argument.
    """

    def __init__(self, scene, dtype=np.int16, band_axis=2,
                 area_aggregation=True,
                 time_dimension=False):
        if not issubclass(dtype, np.integer):
            raise TypeError('Only integer saving allowed for CF data')

        time_axis = -1
        if time_dimension:
            time_axis = 0

        self.info = scene.info.copy()
        if "time" in self.info:
            del self.info["time"]
        CF_DATA_TYPE = dtype

        # Other global attributes
        self.info["Conventions"] = "CF-1.5"
        #self.info["platform"] = scene.satname + "-" + str(scene.number)
        self.info["platform"] = scene.fullname
        self.info["instrument"] = scene.instrument_name
        if scene.variant:
            self.info["service"] = scene.variant
        if scene.orbit:
            self.info["orbit"] = scene.orbit

        self.time = InfoObject()
        self.time.data = date2num(scene.time_slot,
                                  TIME_UNITS)
        if time_dimension:
            var_dim_names = ("time", )
        else:
            var_dim_names = ()

        self.time.info = {"var_name": "time",
                          "var_data": self.time.data,
                          "var_dim_names": var_dim_names,
                          "long_name": "Nominal time of the image",
                          "standard_name": "time",
                          "units": TIME_UNITS}

        grid_mappings = []
        areas = []
        area = None
        area_units = []
        counter = 0
        gm_counter = 0
        area_counter = 0

        for chn in scene:

            if not chn.is_loaded():
                continue

            if not isinstance(chn, Channel):
                setattr(self, chn.name, chn)
                continue

            fill_value = np.iinfo(CF_DATA_TYPE).min

            if ma.count_masked(chn.data) == chn.data.size:
                # All data is masked
                data = np.ones(chn.data.shape, dtype=CF_DATA_TYPE) * fill_value
                scale = 1
                offset = 0
            else:
                chn_max = chn.data.max()
                chn_min = chn.data.min()

                scale = ((chn_max - chn_min) /
                         (2 ** np.iinfo(CF_DATA_TYPE).bits - 2.0))
                # Handle the case where all data has the same value.
                if scale == 0:
                    scale = 1
                if np.iinfo(CF_DATA_TYPE).kind == 'i':
                    # Signed data type
                    offset = (chn_max + chn_min) / 2.0
                else:  # Unsigned data type
                    offset = chn_min - scale

                if isinstance(chn.data, np.ma.MaskedArray):
                    data = (
                        (chn.data.data - offset) / scale).astype(CF_DATA_TYPE)
                    data[chn.data.mask] = fill_value
                else:
                    data = ((chn.data - offset) / scale).astype(CF_DATA_TYPE)

            if time_dimension:
                data = np.ma.expand_dims(data, time_axis)
            elif area_aggregation:
                data = np.ma.expand_dims(data, band_axis)

            # it's a grid mapping
            try:
                if chn.area.proj_dict not in grid_mappings:
                    # create new grid mapping
                    grid_mappings.append(chn.area.proj_dict)
                    area = InfoObject()
                    area.data = 0
                    area.info = {"var_name": "grid_mapping_" + str(gm_counter),
                                 "var_data": area.data,
                                 "var_dim_names": ()}
                    area.info.update(proj2cf(chn.area.proj_dict))
                    area.info.setdefault("units", "m")
                    setattr(self, area.info["var_name"], area)
                    gm_counter += 1
                else:
                    # use an existing grid mapping
                    str_gmc = str(grid_mappings.index(chn.area.proj_dict))
                    area = InfoObject()
                    area.data = 0
                    area.info = {"var_name": "grid_mapping_" + str_gmc,
                                 "var_data": area.data,
                                 "var_dim_names": ()}
                    area.info.update(proj2cf(chn.area.proj_dict))
                    area.info.setdefault("units", "m")

                if(chn.area in areas):
                    str_arc = str(areas.index(chn.area))
                    xy_names = ["y" + str_arc, "x" + str_arc]
                else:
                    areas.append(chn.area)
                    str_arc = str(area_counter)
                    area_counter += 1
                    x__ = InfoObject()
                    chn.area.get_proj_coords(cache=True)
                    x__.data = chn.area.projection_x_coords[0, :]
                    x__.info = {"var_name": "x" + str_arc,
                                "var_data": x__.data,
                                "var_dim_names": ("x" + str_arc,),
                                "units": "rad",
                                "standard_name": "projection_x_coordinate",
                                "long_name": "x coordinate of projection"}
                    if area.info["grid_mapping_name"] == "geostationary":
                        x__.data /= float(
                            area.info["perspective_point_height"])
                        xpix = np.arange(len(x__.data), dtype=np.uint16)
                        xsca = ((x__.data[-1] - x__.data[0]) /
                                (xpix[-1] + xpix[0]))
                        xoff = x__.data[0] - xpix[0] * xsca
                        x__.data = xpix
                        x__.info["var_data"] = xpix
                        x__.info["scale_factor"] = xsca
                        x__.info["add_offset"] = xoff
                    setattr(self, x__.info["var_name"], x__)

                    y__ = InfoObject()
                    y__.data = chn.area.projection_y_coords[:, 0]
                    y__.info = {"var_name": "y" + str_arc,
                                "var_data": y__.data,
                                "var_dim_names": ("y" + str_arc,),
                                "units": "rad",
                                "standard_name": "projection_y_coordinate",
                                "long_name": "y coordinate of projection"}
                    if area.info["grid_mapping_name"] == "geostationary":
                        y__.data /= float(
                            area.info["perspective_point_height"])
                        ypix = np.arange(len(y__.data), dtype=np.uint16)
                        ysca = ((y__.data[-1] - y__.data[0]) /
                                (ypix[-1] + ypix[0]))
                        yoff = y__.data[0] - ypix[0] * ysca
                        y__.data = ypix
                        y__.info["var_data"] = ypix
                        y__.info["scale_factor"] = ysca
                        y__.info["add_offset"] = yoff
                    setattr(self, y__.info["var_name"], y__)

                    xy_names = [y__.info["var_name"], x__.info["var_name"]]

            # It's not a grid mapping, go for lons and lats
            except AttributeError:
                area = None
                if(chn.area in areas):
                    str_arc = str(areas.index(chn.area))
                    coordinates = ("lat" + str_arc + " " + "lon" + str_arc)
                else:
                    areas.append(chn.area)
                    str_arc = str(area_counter)
                    area_counter += 1
                    lons = InfoObject()
                    try:
                        lons.data = chn.area.lons[:]
                    except AttributeError:
                        lons.data = scene.area.lons[:]

                    lons.info = {"var_name": "lon" + str_arc,
                                 "var_data": lons.data,
                                 "var_dim_names": ("y" + str_arc,
                                                   "x" + str_arc),
                                 "_FillValue": lons.data.fill_value,
                                 "units": "degrees east",
                                 "long_name": "longitude coordinate",
                                 "standard_name": "longitude"}
                    if lons.data is not None:
                        setattr(self, lons.info["var_name"], lons)

                    lats = InfoObject()
                    try:
                        lats.data = chn.area.lats[:]
                    except AttributeError:
                        lats.data = scene.area.lats[:]

                    lats.info = {"var_name": "lat" + str_arc,
                                 "var_data": lats.data,
                                 "var_dim_names": ("y" + str_arc,
                                                   "x" + str_arc),
                                 "_FillValue": lats.data.fill_value,
                                 "units": "degrees north",
                                 "long_name": "latitude coordinate",
                                 "standard_name": "latitude"}
                    if lats.data is not None:
                        setattr(self, lats.info["var_name"], lats)

                    if lats.data is not None and lons.data is not None:
                        coordinates = (lats.info["var_name"] + " " +
                                       lons.info["var_name"])
                xy_names = ["y" + str_arc, "x" + str_arc]

            if (area_aggregation and not time_dimension and
                    (chn.area, chn.info['units']) in area_units):

                str_cnt = str(area_units.index((chn.area, chn.info['units'])))
                # area has been used before
                band = getattr(self, "band" + str_cnt)

                # data
                band.data = np.concatenate((band.data, data), axis=band_axis)
                band.info["var_data"] = band.data

                # bandname
                bandname = getattr(self, "bandname" + str_cnt)
                bandname.data = np.concatenate((bandname.data,
                                                np.array([chn.name])))
                bandname.info["var_data"] = bandname.data

                # offset
                off_attr = np.concatenate((off_attr,
                                           np.array([offset])))
                band.info["add_offset"] = off_attr

                # scale
                sca_attr = np.concatenate((sca_attr,
                                           np.array([scale])))
                band.info["scale_factor"] = sca_attr

                # wavelength bounds
                bwl = getattr(self, "wl_bnds" + str_cnt)
                bwl.data = np.vstack((bwl.data,
                                      np.array([chn.wavelength_range[0],
                                                chn.wavelength_range[2]])))
                bwl.info["var_data"] = bwl.data

                # nominal_wavelength
                nwl = getattr(self, "nominal_wavelength" + str_cnt)
                nwl.data = np.concatenate((nwl.data,
                                           np.array([chn.wavelength_range[1]])))
                nwl.info["var_data"] = nwl.data

            else:
                # first encounter of this area and unit
                str_cnt = str(counter)
                counter += 1
                area_units.append((chn.area, chn.info["units"]))

                # data

                band = InfoObject()
                band.data = data
                dim_names = xy_names
                if time_dimension:
                    dim_names.insert(time_axis, 'time')
                elif area_aggregation:
                    dim_names.insert(band_axis, 'band' + str_cnt)

                band.info = {"var_name": "Image" + str_cnt,
                             "var_data": band.data,
                             'var_dim_names': dim_names,
                             "_FillValue": fill_value,
                             "long_name": "Band data",
                             "units": chn.info["units"],
                             "resolution": chn.resolution}

                # bandname

                var_dim_names = ("band" + str_cnt,)
                bandname = InfoObject()
                bandname.data = np.array([chn.name], 'O')
                bandname.info = {"var_name": "band" + str_cnt,
                                 "var_data": bandname.data,
                                 "var_dim_names": var_dim_names,
                                 "standard_name": "band_name"}
                setattr(self, "bandname" + str_cnt, bandname)

                # offset
                off_attr = np.array([offset])
                band.info["add_offset"] = off_attr

                # scale
                sca_attr = np.array([scale])
                band.info["scale_factor"] = sca_attr

                # wavelength bounds
                wlbnds = InfoObject()
                wlbnds.data = np.array([[chn.wavelength_range[0],
                                         chn.wavelength_range[2]]])
                wlbnds.info = {"var_name": "wl_bnds" + str_cnt,
                               "var_data": wlbnds.data,
                               "var_dim_names": ("band" + str_cnt, "nv")}
                setattr(self, wlbnds.info["var_name"], wlbnds)

                # nominal_wavelength
                nomwl = InfoObject()
                nomwl.data = np.array([chn.wavelength_range[1]])
                nomwl.info = {"var_name": "nominal_wavelength" + str_cnt,
                              "var_data": nomwl.data,
                              "var_dim_names": ("band" + str_cnt,),
                              "standard_name": "radiation_wavelength",
                              "units": "um",
                              "bounds": wlbnds.info["var_name"]}
                setattr(self, "nominal_wavelength" + str_cnt, nomwl)

                # grid mapping or lon lats
                if area is not None:
                    band.info["grid_mapping"] = area.info["var_name"]
                else:
                    band.info["coordinates"] = coordinates

                # Add other (custom) attributes:
                # Only scalar attributes!
                for key in chn.info.keys():
                    if key not in band.info.keys():
                        if (type(chn.info[key]) == str or type(chn.info[key]) == int or
                                type(chn.info[key]) == float):
                            band.info[key] = chn.info[key]

                setattr(self, "band" + str_cnt, band)

        for i, area_unit in enumerate(area_units):
            # compute data reduction
            fill_value = np.iinfo(CF_DATA_TYPE).min
            band = getattr(self, "band" + str(i))
            valid_min, valid_max = band.data.min(), band.data.max()
            band.info["valid_range"] = np.array([valid_min, valid_max]),

    def save(self, filename, *args, **kwargs):
        return netcdf_cf_writer(filename, self, kwargs.get("compression", True))


def proj2cf(proj_dict):
    """Return the cf grid mapping from a proj dict.

    Description of the cf grid mapping:
    http://cf-pcmdi.llnl.gov/documents/cf-conventions/1.4/ch05s06.html

    Table of the available grid mappings:
    http://cf-pcmdi.llnl.gov/documents/cf-conventions/1.4/apf.html
    """

    cases = {"geos": geos2cf,
             "stere": stere2cf,
             "merc": merc2cf,
             "aea": aea2cf,
             "laea": laea2cf,
             "ob_tran": obtran2cf,
             "eqc": eqc2cf, }

    return cases[proj_dict["proj"]](proj_dict)


def geos2cf(proj_dict):
    """Return the cf grid mapping from a geos proj dict.
    """

    return {"grid_mapping_name": "geostationary",
            "latitude_of_projection_origin": 0.0,
            "longitude_of_projection_origin": eval(proj_dict["lon_0"]),
            "semi_major_axis": eval(proj_dict["a"]),
            "semi_minor_axis": eval(proj_dict["b"]),
            "perspective_point_height": eval(proj_dict["h"])
            }


def eqc2cf(proj_dict):
    """Return the cf grid mapping from a eqc proj dict. However, please be
    aware that this is not an official CF projection. See
    http://cf-pcmdi.llnl.gov/documents/cf-conventions/1.4/apf.html.
    """

    return {"grid_mapping_name": "equirectangular",
            "latitude_of_true_scale": eval(proj_dict.get("lat_ts", "0")),
            "latitude_of_projection_origin": eval(proj_dict["lat_0"]),
            "longitude_of_projection_origin": eval(proj_dict["lon_0"]),
            "false_easting": eval(proj_dict.get("x_0", "0")),
            "false_northing": eval(proj_dict.get("y_0", "0"))
            }


def stere2cf(proj_dict):
    """Return the cf grid mapping from a stereographic proj dict.
    """

    return {"grid_mapping_name": "stereographic",
            "latitude_of_projection_origin": eval(proj_dict["lat_0"]),
            "longitude_of_projection_origin": eval(proj_dict["lon_0"]),
            "scale_factor_at_projection_origin": eval(
                proj_dict.get("x_0", "1.0")),
            "false_easting": eval(proj_dict.get("x_0", "0")),
            "false_northing": eval(proj_dict.get("y_0", "0"))
            }


def merc2cf(proj_dict):
    """Return the cf grid mapping from a mercator proj dict.
    """

    raise NotImplementedError(
        "CF grid mapping from a PROJ.4 mercator projection is not implemented")


def aea2cf(proj_dict):
    """Return the cf grid mapping from a Albers Equal Area proj dict.
    """

    #standard_parallels = []
    # for item in ['lat_1', 'lat_2']:
    #    if item in proj_dict:
    #        standard_parallels.append(eval(proj_dict[item]))
    if 'lat_2' in proj_dict:
        standard_parallel = [eval(proj_dict['lat_1']),
                             eval(proj_dict['lat_2'])]
    else:
        standard_parallel = [eval(proj_dict['lat_1'])]

    lat_0 = 0.0
    if 'lat_0' in proj_dict:
        lat_0 = eval(proj_dict['lat_0'])

    x_0 = 0.0
    if 'x_0' in proj_dict:
        x_0 = eval(proj_dict['x_0'])

    y_0 = 0.0
    if 'y_0' in proj_dict:
        y_0 = eval(proj_dict['y_0'])

    retv = {"grid_mapping_name": "albers_conical_equal_area",
            "standard_parallel": standard_parallel,
            "latitude_of_projection_origin": lat_0,
            "longitude_of_central_meridian": eval(proj_dict["lon_0"]),
            "false_easting": x_0,
            "false_northing": y_0
            }

    retv = build_dict("albers_conical_equal_area",
                      proj_dict,
                      standard_parallel=["lat_1", "lat_2"],
                      latitude_of_projection_origin="lat_0",
                      longitude_of_central_meridian="lon_0",
                      false_easting="x_0",
                      false_northing="y_0")

    return retv


def laea2cf(proj_dict):
    """Return the cf grid mapping from a Lambert azimuthal equal-area proj dict.
    http://trac.osgeo.org/gdal/wiki/NetCDF_ProjectionTestingStatus
    """
    x_0 = eval(proj_dict.get('x_0', '0.0'))
    y_0 = eval(proj_dict.get('y_0', '0.0'))

    retv = {"grid_mapping_name": "lambert_azimuthal_equal_area",
            "longitude_of_projection_origin": eval(proj_dict["lon_0"]),
            "latitude_of_projection_origin": eval(proj_dict["lat_0"]),
            "false_easting": x_0,
            "false_northing": y_0
            }

    retv = build_dict("lambert_azimuthal_equal_area",
                      proj_dict,
                      longitude_of_projection_origin="lon_0",
                      latitude_of_projection_origin="lat_0",
                      false_easting="x_0",
                      false_northing="y_0")

    return retv


def obtran2cf(proj_dict):
    """Return a grid mapping from a rotated pole grid (General Oblique
    Transformation projection) proj dict.

    Please be aware this is not yet supported by CF!
    """
    LOG.warning("The General Oblique Transformation " +
                "projection is not CF compatible yet...")
    x_0 = eval(proj_dict.get('x_0', '0.0'))
    y_0 = eval(proj_dict.get('y_0', '0.0'))

    retv = {"grid_mapping_name": "general_oblique_transformation",
            "longitude_of_projection_origin": eval(proj_dict["lon_0"]),
            "grid_north_pole_latitude": eval(proj_dict["o_lat_p"]),
            "grid_north_pole_longitude": eval(proj_dict["o_lon_p"]),
            "false_easting": x_0,
            "false_northing": y_0
            }

    retv = build_dict("general_oblique_transformation",
                      proj_dict,
                      longitude_of_projection_origin="lon_0",
                      grid_north_pole_latitude="o_lat_p",
                      grid_north_pole_longitude="o_lon_p",
                      false_easting="x_0",
                      false_northing="y_0")

    return retv


def build_dict(proj_name, proj_dict, **kwargs):
    new_dict = {}
    new_dict["grid_mapping_name"] = proj_name
    for key, val in kwargs.items():
        if isinstance(val, (list, tuple)):
            new_dict[key] = [eval(proj_dict[x]) for x in val if x in proj_dict]
        elif val in proj_dict:
            new_dict[key] = eval(proj_dict[val])
    # add a, b, rf and/or ellps
    if "a" in proj_dict:
        new_dict["semi_major_axis"] = eval(proj_dict["a"])
    if "b" in proj_dict:
        new_dict["semi_minor_axis"] = eval(proj_dict["b"])
    if "rf" in proj_dict:
        new_dict["inverse_flattening"] = eval(proj_dict["rf"])
    if "ellps" in proj_dict:
        new_dict["ellipsoid"] = proj_dict["ellps"]

    return new_dict


def aeqd2cf(proj_dict):
    return build_dict("azimuthal_equidistant",
                      proj_dict,
                      standard_parallel=["lat_1", "lat_2"],
                      latitude_of_projection_origin="lat_0",
                      longitude_of_central_meridian="lon_0",
                      false_easting="x_0",
                      false_northing="y_0")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2010, 2011, 2013, 2014, 2015.

# SMHI,
# Folkborgsvägen 1,
# Norrköping,
# Sweden

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>
#   Esben S. Nielsen <esn@dmi.dk>
#   Panu Lahtinen <panu.lahtinen@fmi.fi>
#   Adam Dybbroe <adam.dybbroe@smhi.se>

# This file is part of mpop.

# mpop is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# mpop is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# mpop.  If not, see <http://www.gnu.org/licenses/>.

"""Interface to Eumetcast level 1.5 HRIT/LRIT format. Uses the MIPP reader.
"""
import ConfigParser
import fnmatch
import logging
import os

from pyproj import Proj

from mipp import CalibrationError, ReaderError, xrit
from mpop import CONFIG_PATH
from mpop.plugin_base import Reader
from mpop.satin.helper_functions import area_def_names_to_extent

LOGGER = logging.getLogger(__name__)

try:
    # Work around for on demand import of pyresample. pyresample depends
    # on scipy.spatial which memory leaks on multiple imports
    IS_PYRESAMPLE_LOADED = False
    from pyresample import geometry
    from mpop.projector import get_area_def
    IS_PYRESAMPLE_LOADED = True
except ImportError:
    LOGGER.warning("pyresample missing. Can only work in satellite projection")


class XritReader(Reader):

    '''Class for reading XRIT data.
    '''
    pformat = "mipp_xrit"

    def load(self, *args, **kwargs):
        load(*args, **kwargs)


def load(satscene, calibrate=True, area_extent=None, area_def_names=None,
         **kwargs):
    """Read data from file and load it into *satscene*. The *calibrate*
    argument is passed to mipp (should be 0 for off, 1 for default, and 2 for
    radiances only).
    """

    conf = ConfigParser.RawConfigParser()
    conf.read(os.path.join(CONFIG_PATH, satscene.fullname + ".cfg"))
    options = {}
    for option, value in conf.items(satscene.instrument_name + "-level2"):
        options[option] = value

    for section in conf.sections():
        if(section.startswith(satscene.instrument_name) and
           not (section == "satellite") and
           # not section[:-1].endswith("-level") and
           not section.endswith("-granules")):
            options[section] = dict(conf.items(section))

    filenames = kwargs.get('filename')

    CASES.get(satscene.instrument_name, load_generic)(satscene,
                                                      options,
                                                      calibrate,
                                                      area_extent,
                                                      area_def_names,
                                                      filenames)


def load_generic(satscene, options, calibrate=True, area_extent=None,
                 area_def_names=None, filenames=None):
    """Read imager data from file and load it into *satscene*.
    """

    os.environ["PPP_CONFIG_DIR"] = CONFIG_PATH

    LOGGER.debug("Channels to load from %s: %s" % (satscene.instrument_name,
                                                   satscene.channels_to_load))

    # Compulsory global attributes
    satscene.info["title"] = (satscene.satname.capitalize() + satscene.number +
                              " satellite, " +
                              satscene.instrument_name.capitalize() +
                              " instrument.")
    satscene.info["institution"] = "Original data disseminated by EumetCast."
    satscene.add_to_history("HRIT/LRIT data read by mipp/mpop.")
    satscene.info["references"] = "No reference."
    satscene.info["comments"] = "No comment."

    from_area = False

    if satscene.end_time is not None:
        time_slot = satscene.time_slot, satscene.end_time
    else:
        time_slot = satscene.time_slot

    if area_extent is None and satscene.area is not None:
        if not satscene.area_def:
            satscene.area = get_area_def(satscene.area_id)
        area_extent = satscene.area.area_extent
        from_area = True

    area_converted_to_extent = False

    for chn in satscene.channels_to_load:
        use_filenames = False
        # Sort out filenames
        if filenames is not None:
            for section in options.keys():
                if section.endswith('-level1'):
                    break
            try:
                pattern_pro = eval(options[section].get('filename_pro'))
            except TypeError:
                pattern_pro = None
            try:
                pattern_epi = eval(options[section].get('filename_epi'))
            except TypeError:
                pattern_epi = None
            pattern = eval(options[section].get('filename'))

            epilogue = None
            prologue = None
            image_files = []

            if pattern_epi is not None:
                glob_epi = satscene.time_slot.strftime(
                    pattern_epi) % ({'segment': "EPI".ljust(9, '_'),
                                     'channel': chn + '*'})
            else:
                glob_epi = 'eggs_and_spam'

            if pattern_pro is not None:
                glob_pro = satscene.time_slot.strftime(
                    pattern_pro) % ({'segment': "PRO".ljust(9, '_'),
                                     'channel': chn + '*'})
            else:
                glob_pro = 'eggs_and_spam'

            glob_img = satscene.time_slot.strftime(
                pattern) % ({'segment': "*", 'channel': chn + '*'})

            for filename in filenames:
                if fnmatch.fnmatch(os.path.basename(filename), glob_img):
                    image_files.append(filename)
                elif pattern_pro is not None and fnmatch.fnmatch(
                    os.path.basename(filename),
                        glob_pro):
                    prologue = filename
                elif pattern_epi is not None and fnmatch.fnmatch(
                    os.path.basename(filename),
                        glob_epi):
                    epilogue = filename
            if len(image_files) == 0 and prologue is None and epilogue is None:
                use_filenames = False
            else:
                use_filenames = True

        if from_area:
            try:
                if use_filenames:
                    metadata = xrit.sat.load_files(prologue,
                                                   image_files,
                                                   epilogue,
                                                   platform_name=satscene.fullname,
                                                   only_metadata=True)
                else:
                    metadata = xrit.sat.load(satscene.fullname,
                                             time_slot,
                                             chn,
                                             only_metadata=True)
                if(satscene.area_def.proj_dict["proj"] != "geos" or
                   float(satscene.area_def.proj_dict["lon_0"]) !=
                   metadata.sublon):
                    raise ValueError("Slicing area must be in "
                                     "geos projection, and lon_0 should match "
                                     "the satellite's position.")
            except ReaderError, err:
                # if channel can't be found, go on with next channel
                LOGGER.error(str(err))
                continue

        # Convert area definitions to maximal area_extent
        if not area_converted_to_extent and area_def_names is not None:
            try:
                if use_filenames:
                    metadata = xrit.sat.load_files(prologue,
                                                   image_files,
                                                   epilogue,
                                                   platform_name=satscene.fullname,
                                                   only_metadata=True)
                else:
                    metadata = xrit.sat.load(satscene.fullname,
                                             time_slot,
                                             chn,
                                             only_metadata=True)
            except ReaderError as err:
                LOGGER.warning(str(err))
                continue
            # if area_extent is given, assume it gives the maximum
            # extent of the satellite view
            if area_extent is not None:
                area_extent = area_def_names_to_extent(area_def_names,
                                                       metadata.proj4_params,
                                                       area_extent)
            # otherwise use the default value (MSG3 extent at
            # lon0=0.0), that is, do not pass default_extent=area_extent
            else:
                area_extent = area_def_names_to_extent(area_def_names,
                                                       metadata.proj4_params,
                                                       default_extent=None)

            if area_extent is None:
                LOGGER.info('Could not derive area_extent from area_def_names')

            area_converted_to_extent = True

        try:
            if use_filenames:
                image = xrit.sat.load_files(prologue,
                                            image_files,
                                            epilogue,
                                            platform_name=satscene.fullname,
                                            mask=True,
                                            calibrate=calibrate)
            else:
                image = xrit.sat.load(satscene.fullname,
                                      time_slot,
                                      chn,
                                      mask=True,
                                      calibrate=calibrate)
            if area_extent:
                metadata, data = image(area_extent)
            else:
                metadata, data = image()
        except CalibrationError:
            LOGGER.warning(
                "Loading non calibrated data since calibration failed.")
            if use_filenames:
                image = xrit.sat.load_files(prologue,
                                            image_files,
                                            epilogue,
                                            platform_name=satscene.fullname,
                                            mask=True,
                                            calibrate=False)
            else:
                image = xrit.sat.load(satscene.fullname,
                                      time_slot,
                                      chn,
                                      mask=True,
                                      calibrate=False)
            if area_extent:
                metadata, data = image(area_extent)
            else:
                metadata, data = image()

        except ReaderError as err:
            # if channel can't be found, go on with next channel
            LOGGER.warning(str(err))
            continue

        satscene[chn] = data

        satscene[chn].info['units'] = metadata.calibration_unit
        satscene[chn].info['sublon'] = metadata.sublon
        satscene[chn].info['satname'] = satscene.satname
        satscene[chn].info['satnumber'] = satscene.number
        satscene[chn].info['instrument_name'] = satscene.instrument_name
        satscene[chn].info['time'] = satscene.time_slot

        # Build an area on the fly from the mipp metadata
        proj_params = getattr(metadata, "proj4_params").split(" ")
        proj_dict = {}
        for param in proj_params:
            key, val = param.split("=")
            proj_dict[key] = val

        if IS_PYRESAMPLE_LOADED:
            # Build area_def on-the-fly
            satscene[chn].area = geometry.AreaDefinition(
                satscene.satname + satscene.instrument_name +
                str(metadata.area_extent) +
                str(data.shape),
                "On-the-fly area",
                proj_dict["proj"],
                proj_dict,
                data.shape[1],
                data.shape[0],
                metadata.area_extent)
        else:
            LOGGER.info("Could not build area, pyresample missing...")

CASES = {}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2010, 2011, 2014.

# Author(s):
 
#   Martin Raspaud <martin.raspaud@smhi.se>

# This file is part of mpop.

# mpop is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.

# mpop is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
# A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

# You should have received a copy of the GNU General Public License along with
# mpop.  If not, see <http://www.gnu.org/licenses/>.

"""Module for testing the :mod:`mpop.instruments.seviri` module.
"""
import random
import unittest

import numpy as np

import mpop.instruments.seviri
from mpop.imageo import geo_image


def patch_scene_mask():
    """Patch the :mod:`mpop.scene` module to avoid using it in these tests.
    """
    class FakeChannel(object):
        """FakeChannel class.
        """
        def __init__(self, val):
            del val
            self.data = np.ma.array((random.random(),))

        def check_range(self, *args):
            """Dummy check_range function.
            """
            del args
            return self.data

    
    class FakeSatscene(object):
        """Fake SatelliteInstrumentScene.
        """
        __version__ = "fake"
        def __init__(self):
            self.channels = {}
            self.area = None
            self.time_slot = None
            self.error = []
        def check_channels(self, *args):
            """Dummy check_channels function.
            """
            for chn in args:
                if chn in self.error:
                    raise RuntimeError()

        def __getitem__(self, key):
            if key not in self.channels:
                self.channels[key] = FakeChannel(key)
            return self.channels[key]
        
    mpop.instruments.visir.OldVisirCompositer = mpop.instruments.visir.VisirCompositer
    mpop.instruments.visir.VisirCompositer = FakeSatscene
    reload(mpop)
    reload(mpop.instruments)
    reload(mpop.instruments.seviri)

def patch_scene():
    """Patch the :mod:`mpop.scene` module to avoid using it in these tests.
    """
    class FakeChannel(object):
        """FakeChannel class.
        """
        def __init__(self, val):
            self.data = val

        def check_range(self, *args):
            """Dummy check_range function.
            """
            del args
            return self.data

    
    class FakeSatscene(object):
        """Fake SatelliteInstrumentScene.
        """
        __version__ = "fake"
        def __init__(self):
            self.channels = None
            self.area = None
            self.time_slot = None
            self._data_holder = self
            
        def check_channels(self, *args):
            """Dummy check_channels function.
            """
            self.channels = args

        def __contains__(self, point):
            return True
            

        def __getitem__(self, key):
            if key == "_IR39Corr":
                return FakeChannel(3.75)
            elif key == "HRV":
                return FakeChannel(0.7)
            return FakeChannel(key)

    mpop.instruments.visir.OldVisirCompositer = mpop.instruments.visir.VisirCompositer
    mpop.instruments.visir.VisirCompositer = FakeSatscene
    reload(mpop)
    reload(mpop.instruments)
    reload(mpop.instruments.seviri)

def unpatch_scene():
    """Unpatch the :mod:`mpop.scene` module.
    """
    mpop.instruments.visir.VisirCompositer = mpop.instruments.visir.OldVisirCompositer
    delattr(mpop.instruments.visir, "OldVisirCompositer")
    reload(mpop)
    reload(mpop.instruments)
    reload(mpop.instruments.visir)
    reload(mpop.instruments.seviri)

def patch_geo_image():
    """Patch the :mod:`imageo.geo_image` module to avoid using it in these
    tests.
    """
    class FakeGeoImage:
        """FakeGeoImage class.
        """
        def __init__(self, *args, **kwargs):
            self.args = args
            self.kwargs = kwargs
            self.lum = None
            self.channels = [self]
            
        def enhance(self, **kwargs):
            """Dummy enhance function.
            """
            self.kwargs.update(kwargs)

        def replace_luminance(self, lum):
            """Dummy remplace_luminance.
            """
            self.lum = lum
        

    geo_image.OldGeoImage = geo_image.GeoImage
    geo_image.GeoImage = FakeGeoImage

def unpatch_geo_image():
    """Unpatch the :mod:`imageo.geo_image` module.
    """
    geo_image.GeoImage = geo_image.OldGeoImage
    delattr(geo_image, "OldGeoImage")

class TestComposites(unittest.TestCase):
    """Class for testing the composites.
    """

    def setUp(self):
        """Setup stuff.
        """
        patch_geo_image()
        patch_scene()
        self.scene = mpop.instruments.seviri.SeviriCompositer()


    # def test_cloudtop(self):
    #     """Test cloudtop.
    #     """
    #     img = self.scene.cloudtop()
    #     self.assertEquals(img.kwargs["mode"], "RGB")
    #     self.assertEquals(img.kwargs["fill_value"], (0, 0, 0))
    #     self.assertEquals(img.args[0], (-3.75, -10.8, -12.0))
    #     self.assertEquals(img.kwargs["stretch"], (0.005, 0.005))
    #     self.assertTrue("crange" not in img.kwargs)
    #     self.assertTrue("gamma" not in img.kwargs)

    # def test_night_fog(self):
    #     """Test night_fog.
    #     """
    #     img = self.scene.night_fog()
    #     self.assertEquals(img.kwargs["mode"], "RGB")
    #     self.assertEquals(img.kwargs["fill_value"], (0, 0, 0))
    #     self.assertEquals(img.args[0], (12.0 - 10.8, 10.8 - 3.75, 10.8))
    #     self.assertEquals(img.kwargs["crange"], ((-4, 2),
    #                                              (0, 6),
    #                                              (243, 293)))
    #     self.assertEquals(img.kwargs["gamma"], (1.0, 2.0, 1.0))
    #     self.assertTrue("stretch" not in img.kwargs)

#     def test_hr_overview(self):
#         """Test hr_overview.
#         """
#         img = self.scene.hr_overview()
#         self.assertEquals(img.kwargs["mode"], "RGB")
#         self.assertEquals(img.kwargs["fill_value"], (0, 0, 0))
#         self.assertEquals(img.args[0], (0.635, 0.85, -10.8))
#         self.assertEquals(img.kwargs["stretch"], "crude")
#         self.assertEquals(list(img.kwargs["gamma"]), list((1.6, 1.6, 1.1)))
#         self.assertTrue("crange" not in img.kwargs)

#         self.assertEquals(img.lum.kwargs["mode"], "L")
#         self.assertEquals(img.lum.kwargs["crange"], (0, 100))
#         self.assertEquals(img.lum.kwargs["gamma"], 2.0)
#         self.assertTrue("stretch" not in img.lum.kwargs)
#         self.assertTrue("fill_value" not in img.lum.kwargs)

#     def test_hr_visual(self):
#         """Test hr_visual.
#         """
#         img = self.scene.hr_visual()
#         self.assertEquals(img.kwargs["mode"], "L")
#         self.assertEquals(img.kwargs["fill_value"], 0)
#         self.assertEquals(img.args[0], 0.7)
#         self.assertEquals(img.kwargs["stretch"], "crude")
#         self.assertTrue("crange" not in img.kwargs)
#         self.assertTrue("gamma" not in img.kwargs)
        
        

    def tearDown(self):
        unpatch_scene()
        unpatch_geo_image()


class TestCo2Corr(unittest.TestCase):
    """Class for testing the composites.
    """

    def setUp(self):
        """Setup stuff.
        """
        patch_geo_image()
        patch_scene_mask()
        self.scene = mpop.instruments.seviri.SeviriCompositer()


    def test_co2corr(self):
        """Test CO2 correction.
        """
        res = self.scene.co2corr()
        bt039 = self.scene[3.9].data
        bt108 = self.scene[10.8].data
        bt134 = self.scene[13.4].data
        
        dt_co2 = (bt108-bt134)/4.0
        rcorr = bt108 ** 4 - (bt108-dt_co2) ** 4
        
        
        t4_co2corr = bt039 ** 4 + rcorr
        if t4_co2corr < 0.0:
            t4_co2corr = 0
        solution = t4_co2corr ** 0.25
        self.assertEquals(res, solution)

        self.scene.error = [3.75]
        res = self.scene.co2corr()

        self.assertTrue(res is None)


    def tearDown(self):
        unpatch_scene()
        unpatch_geo_image()


def suite():
    """The test suite for test_seviri.
    """
    loader = unittest.TestLoader()
    mysuite = unittest.TestSuite()
    mysuite.addTest(loader.loadTestsFromTestCase(TestComposites))
    mysuite.addTest(loader.loadTestsFromTestCase(TestCo2Corr))
    
    return mysuite



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright (c) 2009, 2014.

# SMHI,
# Folkborgsvägen 1,
# Norrköping,
# Sweden

# Author(s):

#   Martin Raspaud <martin.raspaud@smhi.se>
#   Adam Dybbroe <adam.dybbroe@smhi.se>

# This file is part of the mpop.

# mpop is free software: you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# mpop is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with mpop.  If not, see <http://www.gnu.org/licenses/>.

"""Saturn Package initializer.
"""
